# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
from __future__ import annotations

import time
from collections.abc import Callable
from typing import Any

import aqt
import aqt.forms
import aqt.main
from anki.decks import DeckId
from anki.utils import is_mac
from aqt import gui_hooks
from aqt.operations.deck import set_current_deck
from aqt.qt import *
from aqt.theme import theme_manager
from aqt.utils import (
    add_close_shortcut,
    disable_help_button,
    getSaveFile,
    maybeHideClose,
    restoreGeom,
    saveGeom,
    tooltip,
    tr,
)
from aqt.webview import LegacyStatsWebView


class NewDeckStats(QDialog):
    """New deck stats."""

    def __init__(self, mw: aqt.main.AnkiQt) -> None:
        QDialog.__init__(self, mw, Qt.WindowType.Window)
        mw.garbage_collect_on_dialog_finish(self)
        self.mw = mw
        self.name = "deckStats"
        self.period = 0
        self.form = aqt.forms.stats.Ui_Dialog()
        self.oldPos = None
        self.wholeCollection = False
        self.setMinimumWidth(700)
        disable_help_button(self)
        f = self.form
        f.setupUi(self)
        f.groupBox.setVisible(False)
        f.groupBox_2.setVisible(False)
        if not is_mac:
            f.horizontalLayout_4.setContentsMargins(0, 0, 0, 0)
        restoreGeom(self, self.name, default_size=(800, 800))

        from aqt.deckchooser import DeckChooser

        self.deck_chooser = DeckChooser(
            self.mw,
            f.deckArea,
            on_deck_changed=self.on_deck_changed,
            dyn=True,  # include filtered decks
        )

        b = f.buttonBox.addButton(
            tr.statistics_save_pdf(), QDialogButtonBox.ButtonRole.ActionRole
        )
        assert b is not None
        qconnect(b.clicked, self.saveImage)
        b.setAutoDefault(False)
        b = f.buttonBox.button(QDialogButtonBox.StandardButton.Close)
        assert b is not None
        b.setAutoDefault(False)
        maybeHideClose(self.form.buttonBox)
        add_close_shortcut(self)
        gui_hooks.stats_dialog_will_show(self)
        self.form.web.hide_while_preserving_layout()
        self.show()
        self.refresh()
        self.form.web.set_bridge_command(self._on_bridge_cmd, self)
        self.activateWindow()

    def reject(self) -> None:
        self.deck_chooser.cleanup()
        self.form.web.cleanup()
        self.form.web = None  # type: ignore
        saveGeom(self, self.name)
        aqt.dialogs.markClosed("NewDeckStats")
        QDialog.reject(self)

    def closeWithCallback(self, callback: Callable[[], None]) -> None:
        self.reject()
        callback()

    def on_deck_changed(self, deck_id: int) -> None:
        set_current_deck(parent=self, deck_id=DeckId(deck_id)).success(
            lambda _: self.refresh()
        ).run_in_background()

    def _imagePath(self) -> str | None:
        name = time.strftime("-%Y-%m-%d@%H-%M-%S.pdf", time.localtime(time.time()))
        name = f"anki-{tr.statistics_stats()}{name}"
        file = getSaveFile(
            self,
            title=tr.statistics_save_pdf(),
            dir_description="stats",
            key="stats",
            ext=".pdf",
            fname=name,
        )
        return file

    def saveImage(self) -> None:
        path = self._imagePath()
        if not path:
            return

        # When scrolled down in dark mode, the top of the page in the
        # final PDF will have a white background, making the text and graphs
        # unreadable. A simple fix for now is to scroll to the top of the
        # page first.
        def after_scroll(arg: Any) -> None:
            form_web_page = self.form.web.page()
            assert form_web_page is not None
            form_web_page.printToPdf(path)
            tooltip(tr.statistics_saved())

        self.form.web.evalWithCallback("window.scrollTo(0, 0);", after_scroll)

    # legacy add-ons
    def changePeriod(self, n: Any) -> None:
        pass

    def changeScope(self, type: Any) -> None:
        pass

    def _on_bridge_cmd(self, cmd: str) -> bool:
        if cmd.startswith("browserSearch"):
            _, query = cmd.split(":", 1)
            browser = aqt.dialogs.open("Browser", self.mw)
            browser.search_for(query)

        return False

    def refresh(self) -> None:
        self.form.web.load_sveltekit_page("graphs")


class DeckStats(QDialog):
    """Legacy deck stats, used by some add-ons."""

    def __init__(self, mw: aqt.main.AnkiQt) -> None:
        QDialog.__init__(self, mw, Qt.WindowType.Window)
        mw.garbage_collect_on_dialog_finish(self)
        self.mw = mw
        self.name = "deckStats"
        self.period = 0
        self.form = aqt.forms.stats.Ui_Dialog()
        # Hack: Switch out web views dynamically to avoid maintaining multiple
        # Qt forms for different versions of the stats dialog.
        self.form.web = LegacyStatsWebView(self.mw)
        self.oldPos = None
        self.wholeCollection = False
        self.setMinimumWidth(700)
        disable_help_button(self)
        f = self.form
        if theme_manager.night_mode and not theme_manager.macos_dark_mode():
            # the grouping box renders incorrectly in the fusion theme. 5.9+
            # 5.13 behave differently to 5.14, but it looks bad in either case,
            # and adjusting the top margin makes the 'save PDF' button show in
            # the wrong place, so for now we just disable the border instead
            self.setStyleSheet("QGroupBox { border: 0; }")
        f.setupUi(self)
        restoreGeom(self, self.name)
        b = f.buttonBox.addButton(
            tr.statistics_save_pdf(), QDialogButtonBox.ButtonRole.ActionRole
        )
        assert b is not None
        qconnect(b.clicked, self.saveImage)
        b.setAutoDefault(False)
        qconnect(f.groups.clicked, lambda: self.changeScope("deck"))
        f.groups.setShortcut("g")
        qconnect(f.all.clicked, lambda: self.changeScope("collection"))
        qconnect(f.month.clicked, lambda: self.changePeriod(0))
        qconnect(f.year.clicked, lambda: self.changePeriod(1))
        qconnect(f.life.clicked, lambda: self.changePeriod(2))
        maybeHideClose(self.form.buttonBox)
        add_close_shortcut(self)
        gui_hooks.stats_dialog_old_will_show(self)
        self.show()
        self.refresh()
        self.activateWindow()

    def reject(self) -> None:
        self.form.web.cleanup()
        self.form.web = None  # type: ignore
        saveGeom(self, self.name)
        aqt.dialogs.markClosed("DeckStats")
        QDialog.reject(self)

    def closeWithCallback(self, callback: Callable[[], None]) -> None:
        self.reject()
        callback()

    def _imagePath(self) -> str | None:
        name = time.strftime("-%Y-%m-%d@%H-%M-%S.pdf", time.localtime(time.time()))
        name = f"anki-{tr.statistics_stats()}{name}"
        file = getSaveFile(
            self,
            title=tr.statistics_save_pdf(),
            dir_description="stats",
            key="stats",
            ext=".pdf",
            fname=name,
        )
        return file

    def saveImage(self) -> None:
        path = self._imagePath()
        if not path:
            return
        form_web_page = self.form.web.page()
        assert form_web_page is not None
        form_web_page.printToPdf(path)
        tooltip(tr.statistics_saved())

    def changePeriod(self, n: int) -> None:
        self.period = n
        self.refresh()

    def changeScope(self, type: str) -> None:
        self.wholeCollection = type == "collection"
        self.refresh()

    def refresh(self) -> None:
        self.mw.progress.start(parent=self)
        stats = self.mw.col.stats()
        stats.wholeCollection = self.wholeCollection
        self.report = stats.report(type=self.period)
        self.form.web.stdHtml(
            f"<html><body>{self.report}</body></html>",
            js=["js/vendor/jquery.min.js", "js/vendor/plot.js"],
            context=self,
        )
        self.mw.progress.finish()



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import html
from typing import Any

import anki
import anki.find
import aqt
import aqt.forms
from anki.collection import SearchNode
from anki.notes import NoteId
from aqt.qt import *
from aqt.qt import sip

from ..operations import QueryOp
from ..operations.tag import add_tags_to_notes
from ..utils import (
    disable_help_button,
    restore_combo_history,
    restore_combo_index_for_session,
    restoreGeom,
    save_combo_history,
    save_combo_index_for_session,
    saveGeom,
    tr,
)
from . import Browser


class FindDuplicatesDialog(QDialog):
    def __init__(self, browser: Browser, mw: aqt.AnkiQt):
        super().__init__(parent=browser)
        self.browser = browser
        self.mw = mw
        self.mw.garbage_collect_on_dialog_finish(self)
        self.form = form = aqt.forms.finddupes.Ui_Dialog()
        form.setupUi(self)
        restoreGeom(self, "findDupes")
        disable_help_button(self)
        searchHistory = restore_combo_history(form.search, "findDupesFind")

        fields = sorted(
            anki.find.fieldNames(self.mw.col, downcase=False), key=lambda x: x.lower()
        )
        form.fields.addItems(fields)
        restore_combo_index_for_session(form.fields, fields, "findDupesFields")
        self._dupesButton: QPushButton | None = None
        self._dupes: list[tuple[str, list[NoteId]]] = []

        # links
        form.webView.set_bridge_command(self._on_duplicate_clicked, context=self)
        form.webView.stdHtml("", context=self)

        def on_finished(code: Any) -> None:
            saveGeom(self, "findDupes")

        qconnect(self.finished, on_finished)

        def on_click() -> None:
            search_text = save_combo_history(
                form.search, searchHistory, "findDupesFind"
            )
            save_combo_index_for_session(form.fields, "findDupesFields")
            field = fields[form.fields.currentIndex()]
            QueryOp(
                parent=self.browser,
                op=lambda col: col.find_dupes(field, search_text),
                success=self.show_duplicates_report,
            ).run_in_background()

        search = form.buttonBox.addButton(
            tr.actions_search(), QDialogButtonBox.ButtonRole.ActionRole
        )

        assert search is not None

        qconnect(search.clicked, on_click)
        self.show()

    def show_duplicates_report(self, dupes: list[tuple[str, list[NoteId]]]) -> None:
        if sip.isdeleted(self):
            return
        self._dupes = dupes
        if not self._dupesButton:
            self._dupesButton = b = self.form.buttonBox.addButton(
                tr.browsing_tag_duplicates(), QDialogButtonBox.ButtonRole.ActionRole
            )

            assert b is not None

            qconnect(b.clicked, self._tag_duplicates)
        text = ""
        groups = len(dupes)
        notes = sum(len(r[1]) for r in dupes)
        part1 = tr.browsing_group(count=groups)
        part2 = tr.browsing_note_count(count=notes)
        text += tr.browsing_found_as_across_bs(part=part1, whole=part2)
        text += "<p><ol>"
        for val, nids in dupes:
            text += (
                """<li><a href=# onclick="pycmd('%s');return false;">%s</a>: %s</a>"""
                % (
                    html.escape(
                        self.mw.col.build_search_string(
                            SearchNode(nids=SearchNode.IdList(ids=nids))
                        )
                    ),
                    tr.browsing_note_count(count=len(nids)),
                    html.escape(val),
                )
            )
        text += "</ol>"
        self.form.webView.stdHtml(text, context=self)

    def _tag_duplicates(self) -> None:
        if not self._dupes:
            return

        note_ids = set()
        for _, nids in self._dupes:
            note_ids.update(nids)

        add_tags_to_notes(
            parent=self,
            note_ids=list(note_ids),
            space_separated_tags=tr.browsing_duplicate(),
        ).run_in_background()

    def _on_duplicate_clicked(self, link: str) -> None:
        self.browser.search_for(link)
        self.browser.onNote()



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import logging
import sys
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path
from typing import Optional, cast

# All loggers with the following prefix will be treated as add-on loggers
#
# To instatiate a logger with this prefix, use aqt.AddonManager.get_logger()
#
# NOTE: Add-ons might also directly instantiate a logger with this prefix, e.g. in
#       order to avoid depending on the Anki codebase, so this prefix should not
#       be changed.
ADDON_LOGGER_PREFIX = "addon."

# Formatter used for all loggers
FORMATTER = logging.Formatter("%(asctime)s:%(levelname)s:%(name)s: %(message)s")


class AnkiLoggerManager(logging.Manager):
    # inspired by: https://github.com/abdnh/ankiutils/blob/master/src/ankiutils/log.py

    def __init__(
        self,
        logs_path: Path | str,
        existing_loggers: dict[str, logging.Logger | logging.PlaceHolder],
        rootnode: logging.RootLogger,
    ):
        super().__init__(rootnode)
        self.loggerDict = existing_loggers
        self.logs_path = Path(logs_path)

    def getLogger(self, name: str) -> logging.Logger:
        if not name.startswith(ADDON_LOGGER_PREFIX) or name in self.loggerDict:
            return super().getLogger(name)

        # Create a new add-on logger
        logger = super().getLogger(name)

        module = name.split(ADDON_LOGGER_PREFIX)[1].partition(".")[0]
        path = get_addon_logs_folder(self.logs_path, module=module) / f"{module}.log"
        path.parent.mkdir(parents=True, exist_ok=True)

        # Keep the last 10 days of logs
        handler = TimedRotatingFileHandler(
            filename=path, when="D", interval=1, backupCount=10, encoding="utf-8"
        )
        handler.setFormatter(FORMATTER)

        logger.addHandler(handler)

        return logger


def get_addon_logs_folder(logs_path: Path | str, module: str) -> Path:
    return Path(logs_path) / "addons" / module


def find_addon_logger(module: str) -> logging.Logger | None:
    return cast(
        Optional[logging.Logger],
        logging.Logger.manager.loggerDict.get(f"{ADDON_LOGGER_PREFIX}{module}"),
    )


def setup_logging(path: Path | str, **kwargs) -> None:
    """
    Set up logging for the application.

    Configures the root logger to output logs to stdout by default, with custom
    handling for add-on logs. The add-on logs are saved to a separate folder and file
    for each add-on, under the path provided.

    Args:
        path (Path): The path where the log files should be stored.
        **kwargs: Arbitrary keyword arguments for logging.basicConfig
    """

    # Patch root logger manager to handle add-on loggers
    logger_manager = AnkiLoggerManager(
        path, existing_loggers=logging.Logger.manager.loggerDict, rootnode=logging.root
    )
    logging.Logger.manager = logger_manager

    stdout_handler = logging.StreamHandler(stream=sys.stdout)
    stdout_handler.setFormatter(FORMATTER)
    logging.basicConfig(handlers=[stdout_handler], force=True, **kwargs)
    logging.captureWarnings(True)

    # Silence some loggers of external libraries:
    silenced_loggers = [
        "waitress.queue",
    ]
    for logger in silenced_loggers:
        logging.getLogger(logger).setLevel(logging.CRITICAL)
        logging.getLogger(logger).propagate = False



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import json
import re
import time
from collections.abc import Callable
from typing import Any

import aqt.browser
from anki.cards import Card
from anki.collection import Config
from anki.tags import MARKED_TAG
from aqt import AnkiQt, gui_hooks, is_mac
from aqt.qt import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QKeySequence,
    QShortcut,
    Qt,
    QTimer,
    QVBoxLayout,
    qconnect,
)
from aqt.reviewer import replay_audio
from aqt.sound import av_player, play_clicked_audio
from aqt.theme import theme_manager
from aqt.utils import disable_help_button, restoreGeom, saveGeom, setWindowIcon, tr
from aqt.webview import AnkiWebView, AnkiWebViewKind

LastStateAndMod = tuple[str, int, int]


class Previewer(QDialog):
    _last_state: LastStateAndMod | None = None
    _card_changed = False
    _last_render: int | float = 0
    _timer: QTimer | None = None
    _show_both_sides = False

    def __init__(
        self,
        parent: aqt.browser.Browser | None,
        mw: AnkiQt,
        on_close: Callable[[], None],
    ) -> None:
        super().__init__(None, Qt.WindowType.Window)
        mw.garbage_collect_on_dialog_finish(self)
        self._open = True
        self._parent = parent
        self._close_callback = on_close
        self.mw = mw
        disable_help_button(self)
        setWindowIcon(self)
        gui_hooks.previewer_did_init(self)

    def card(self) -> Card | None:
        raise NotImplementedError

    def card_changed(self) -> bool:
        raise NotImplementedError

    def open(self) -> None:
        self._state = "question"
        self._last_state = None
        self._create_gui()
        self._setup_web_view()
        self.render_card()
        restoreGeom(self, "preview")
        self.show()

    def _create_gui(self) -> None:
        self.setWindowTitle(tr.actions_preview())

        self.close_shortcut = QShortcut(QKeySequence("Ctrl+Shift+P"), self)
        qconnect(self.close_shortcut.activated, self.close)

        qconnect(self.finished, self._on_finished)
        self.silentlyClose = True
        self.vbox = QVBoxLayout()
        spacing = 6
        self.vbox.setContentsMargins(0, 0, 0, 0)
        self.vbox.setSpacing(spacing)
        self._web: AnkiWebView | None = AnkiWebView(kind=AnkiWebViewKind.PREVIEWER)
        self.vbox.addWidget(self._web)
        self.bbox = QDialogButtonBox()
        self.bbox.setContentsMargins(
            spacing, spacing if is_mac else 0, spacing, spacing
        )
        self.bbox.setLayoutDirection(Qt.LayoutDirection.LeftToRight)

        gui_hooks.card_review_webview_did_init(self._web, AnkiWebViewKind.PREVIEWER)

        self._replay = self.bbox.addButton(
            tr.actions_replay_audio(), QDialogButtonBox.ButtonRole.ActionRole
        )
        assert self._replay is not None
        self._replay.setAutoDefault(False)
        self._replay.setShortcut(QKeySequence("R"))
        self._replay.setToolTip(tr.actions_shortcut_key(val="R"))
        qconnect(self._replay.clicked, self._on_replay_audio)

        both_sides_button = QCheckBox(tr.qt_misc_back_side_only())
        both_sides_button.setShortcut(QKeySequence("B"))
        both_sides_button.setToolTip(tr.actions_shortcut_key(val="B"))
        self.bbox.addButton(both_sides_button, QDialogButtonBox.ButtonRole.ActionRole)
        self._show_both_sides = self.mw.col.get_config_bool(
            Config.Bool.PREVIEW_BOTH_SIDES
        )
        both_sides_button.setChecked(self._show_both_sides)
        qconnect(both_sides_button.toggled, self._on_show_both_sides)

        self.vbox.addWidget(self.bbox)
        self.setLayout(self.vbox)

    def _on_finished(self, ok: int) -> None:
        saveGeom(self, "preview")
        self._on_close()

    def _on_replay_audio(self) -> None:
        assert self._web is not None
        card = self.card()
        assert card is not None

        gui_hooks.audio_will_replay(self._web, card, self._state == "question")

        if self._state == "question":
            replay_audio(card, True)
        elif self._state == "answer":
            replay_audio(card, False)

    def _on_close(self) -> None:
        self._open = False
        self._close_callback()

        assert self._web is not None

        self._web.cleanup()
        self._web = None

    def _setup_web_view(self) -> None:
        assert self._web is not None

        self._web.stdHtml(
            self.mw.reviewer.revHtml(),
            css=["css/reviewer.css"],
            js=[
                "js/mathjax.js",
                "js/vendor/mathjax/tex-chtml-full.js",
                "js/reviewer.js",
            ],
            context=self,
        )
        self._web.allow_drops = True
        self._web.eval("_blockDefaultDragDropBehavior();")
        self._web.set_bridge_command(self._on_bridge_cmd, self)

    def _on_bridge_cmd(self, cmd: str) -> Any:
        if cmd.startswith("play:"):
            card = self.card()
            assert card is not None

            play_clicked_audio(cmd, card)

    def _update_flag_and_mark_icons(self, card: Card | None) -> None:
        if card:
            flag = card.user_flag()
            marked = card.note(reload=True).has_tag(MARKED_TAG)
        else:
            flag = 0
            marked = False

        assert self._web is not None

        self._web.eval(f"_drawFlag({flag}); _drawMark({json.dumps(marked)});")

    def render_card(self) -> None:
        self.cancel_timer()
        # Keep track of whether render() has ever been called
        # with cardChanged=True since the last successful render
        self._card_changed |= self.card_changed()
        # avoid rendering in quick succession
        elap_ms = int((time.time() - self._last_render) * 1000)
        delay = 300
        if elap_ms < delay:
            self._timer = self.mw.progress.timer(
                delay - elap_ms, self._render_scheduled, False, parent=self
            )
        else:
            self._render_scheduled()

    def cancel_timer(self) -> None:
        if self._timer:
            self._timer.stop()
            self._timer = None

    def _render_scheduled(self) -> None:
        self.cancel_timer()
        self._last_render = time.time()

        if not self._open:
            return
        c = self.card()
        self._update_flag_and_mark_icons(c)
        func = "_showQuestion"
        ans_txt = ""
        if not c:
            txt = tr.qt_misc_please_select_1_card()
            bodyclass = ""
            self._last_state = None
        else:
            if self._show_both_sides:
                self._state = "answer"
            elif self._card_changed:
                self._state = "question"

            currentState = self._state_and_mod()
            if currentState == self._last_state:
                # nothing has changed, avoid refreshing
                return

            # need to force reload even if answer
            txt = c.question(reload=True)
            ans_txt = c.answer()

            if self._state == "answer":
                func = "_showAnswer"
                txt = ans_txt
            txt = re.sub(r"\[\[type:[^]]+\]\]", "", txt)

            bodyclass = theme_manager.body_classes_for_card_ord(c.ord)

            assert self._web is not None

            if c.autoplay():
                self._web.setPlaybackRequiresGesture(False)
                if self._show_both_sides:
                    # if we're showing both sides at once, remove any audio
                    # from the answer that's appeared on the question already
                    question_audio = c.question_av_tags()
                    only_on_answer_audio = [
                        x for x in c.answer_av_tags() if x not in question_audio
                    ]
                    audio = question_audio + only_on_answer_audio
                elif self._state == "question":
                    audio = c.question_av_tags()
                else:
                    audio = c.answer_av_tags()
            else:
                audio = []
                self._web.setPlaybackRequiresGesture(True)
            gui_hooks.av_player_will_play_tags(audio, self._state, self)
            av_player.play_tags(audio)
            txt = self.mw.prepare_card_text_for_display(txt)
            txt = gui_hooks.card_will_show(txt, c, f"preview{self._state.capitalize()}")
            self._last_state = self._state_and_mod()

        js: str
        if self._state == "question":
            ans_txt = self.mw.col.media.escape_media_filenames(ans_txt)
            js = f"{func}({json.dumps(txt)}, {json.dumps(ans_txt)}, '{bodyclass}');"
        else:
            js = f"{func}({json.dumps(txt)}, '{bodyclass}');"

        assert self._web is not None
        self._web.eval(js)
        self._card_changed = False

    def _on_show_both_sides(self, toggle: bool) -> None:
        assert self._web is not None

        self._show_both_sides = toggle
        self.mw.col.set_config_bool(Config.Bool.PREVIEW_BOTH_SIDES, toggle)

        card = self.card()
        assert card is not None

        gui_hooks.previewer_will_redraw_after_show_both_sides_toggled(
            self._web, card, self._state == "question", toggle
        )

        if self._state == "answer" and not toggle:
            self._state = "question"
        self.render_card()

    def _state_and_mod(self) -> tuple[str, int, int]:
        c = self.card()

        assert c is not None

        n = c.note()
        n.load()
        return (self._state, c.id, n.mod)

    def state(self) -> str:
        return self._state


class MultiCardPreviewer(Previewer):
    def card(self) -> Card | None:
        # need to state explicitly it's not implement to avoid W0223
        raise NotImplementedError

    def card_changed(self) -> bool:
        # need to state explicitly it's not implement to avoid W0223
        raise NotImplementedError

    def _create_gui(self) -> None:
        super()._create_gui()
        self._prev = self.bbox.addButton(
            ">" if self.layoutDirection() == Qt.LayoutDirection.RightToLeft else "<",
            QDialogButtonBox.ButtonRole.ActionRole,
        )

        assert self._prev is not None

        self._prev.setAutoDefault(False)
        self._prev.setShortcut(QKeySequence("Left"))
        self._prev.setToolTip(tr.qt_misc_shortcut_key_left_arrow())

        self._next = self.bbox.addButton(
            "<" if self.layoutDirection() == Qt.LayoutDirection.RightToLeft else ">",
            QDialogButtonBox.ButtonRole.ActionRole,
        )

        assert self._next is not None

        self._next.setAutoDefault(True)
        self._next.setShortcut(QKeySequence("Right"))
        self._next.setToolTip(tr.qt_misc_shortcut_key_right_arrow_or_enter())

        qconnect(self._prev.clicked, self._on_prev)
        qconnect(self._next.clicked, self._on_next)

    def _on_prev(self) -> None:
        if self._state == "answer" and not self._show_both_sides:
            self._state = "question"
            self.render_card()
        else:
            self._on_prev_card()

    def _on_prev_card(self) -> None:
        pass

    def _on_next(self) -> None:
        if self._state == "question":
            self._state = "answer"
            self.render_card()
        else:
            self._on_next_card()

    def _on_next_card(self) -> None:
        pass

    def _updateButtons(self) -> None:
        if not self._open:
            return

        assert self._prev is not None
        assert self._next is not None

        self._prev.setEnabled(self._should_enable_prev())
        self._next.setEnabled(self._should_enable_next())

    def _should_enable_prev(self) -> bool:
        return self._state == "answer" and not self._show_both_sides

    def _should_enable_next(self) -> bool:
        return self._state == "question"

    def _on_close(self) -> None:
        super()._on_close()
        self._prev = None
        self._next = None


class BrowserPreviewer(MultiCardPreviewer):
    _last_card_id = 0
    _parent: aqt.browser.Browser | None

    def __init__(
        self, parent: aqt.browser.Browser, mw: AnkiQt, on_close: Callable[[], None]
    ) -> None:
        super().__init__(parent=parent, mw=mw, on_close=on_close)

    def card(self) -> Card | None:
        assert self._parent is not None

        if self._parent.singleCard:
            return self._parent.card
        else:
            return None

    def card_changed(self) -> bool:
        c = self.card()
        if not c:
            return True
        else:
            changed = c.id != self._last_card_id
            self._last_card_id = c.id
            return changed

    def _on_prev_card(self) -> None:
        assert self._parent is not None

        self._parent.onPreviousCard()

    def _on_next_card(self) -> None:
        assert self._parent is not None

        self._parent.onNextCard()

    def _should_enable_prev(self) -> bool:
        assert self._parent is not None

        return super()._should_enable_prev() or self._parent.has_previous_card()

    def _should_enable_next(self) -> bool:
        assert self._parent is not None

        return super()._should_enable_next() or self._parent.has_next_card()

    def _render_scheduled(self) -> None:
        super()._render_scheduled()
        self._updateButtons()



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import html
from collections.abc import Sequence

from anki.collection import OpChanges, OpChangesWithCount, OpChangesWithId
from anki.decks import DeckCollapseScope, DeckDict, DeckId, UpdateDeckConfigs
from aqt.operations import CollectionOp
from aqt.qt import QWidget
from aqt.utils import getOnlyText, tooltip, tr


def remove_decks(
    *,
    parent: QWidget,
    deck_ids: Sequence[DeckId],
    deck_name: str,
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(parent, lambda col: col.decks.remove(deck_ids)).success(
        lambda out: tooltip(
            tr.browsing_cards_deleted_with_deckname(
                count=out.count,
                deck_name=html.escape(deck_name),
            ),
            parent=parent,
        )
    )


def reparent_decks(
    *, parent: QWidget, deck_ids: Sequence[DeckId], new_parent: DeckId
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(
        parent, lambda col: col.decks.reparent(deck_ids=deck_ids, new_parent=new_parent)
    ).success(
        lambda out: tooltip(
            tr.browsing_reparented_decks(count=out.count), parent=parent
        )
    )


def rename_deck(
    *,
    parent: QWidget,
    deck_id: DeckId,
    new_name: str,
) -> CollectionOp[OpChanges]:
    return CollectionOp(
        parent,
        lambda col: col.decks.rename(deck_id, new_name),
    )


def add_deck_dialog(
    *,
    parent: QWidget,
    default_text: str = "",
) -> CollectionOp[OpChangesWithId] | None:
    if name := getOnlyText(
        tr.decks_new_deck_name(),
        default=default_text,
        parent=parent,
        title=tr.decks_create_deck(),
    ).strip():
        return add_deck(parent=parent, name=name)
    else:
        return None


def add_deck(*, parent: QWidget, name: str) -> CollectionOp[OpChangesWithId]:
    return CollectionOp(parent, lambda col: col.decks.add_normal_deck_with_name(name))


def set_deck_collapsed(
    *,
    parent: QWidget,
    deck_id: DeckId,
    collapsed: bool,
    scope: DeckCollapseScope.V,
) -> CollectionOp[OpChanges]:
    return CollectionOp(
        parent,
        lambda col: col.decks.set_collapsed(
            deck_id=deck_id, collapsed=collapsed, scope=scope
        ),
    )


def set_current_deck(*, parent: QWidget, deck_id: DeckId) -> CollectionOp[OpChanges]:
    return CollectionOp(parent, lambda col: col.decks.set_current(deck_id))


def update_deck_configs(
    *, parent: QWidget, input: UpdateDeckConfigs
) -> CollectionOp[OpChanges]:
    return CollectionOp(parent, lambda col: col.decks.update_deck_configs(input))


def update_deck_dict(*, parent: QWidget, deck: DeckDict) -> CollectionOp[OpChanges]:
    return CollectionOp(parent, lambda col: col.decks.update_dict(deck))



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

# make sure not to optimize imports on this file
# ruff: noqa: F401
from __future__ import annotations

import os
import sys
import traceback
from collections.abc import Callable
from typing import TypeVar, Union

from anki._legacy import deprecated

# legacy code depends on these re-exports
from anki.utils import is_mac, is_win

from .qt6 import *


def debug() -> None:
    from pdb import set_trace

    pyqtRemoveInputHook()
    set_trace()


if os.environ.get("DEBUG"):

    def info(type, value, tb) -> None:  # type: ignore
        for line in traceback.format_exception(type, value, tb):
            sys.stdout.write(line)
        pyqtRemoveInputHook()
        from pdb import pm

        pm()

    sys.excepthook = info

_version = QLibraryInfo.version()
qtmajor = _version.majorVersion()
qtminor = _version.minorVersion()
qtpoint = _version.microVersion()
qtfullversion = _version.segments()

if qtmajor == 6 and qtminor < 2:
    raise Exception("Anki does not support your Qt version.")


def qconnect(signal: Callable | pyqtSignal | pyqtBoundSignal, func: Callable) -> None:
    """Helper to work around type checking not working with signal.connect(func)."""
    signal.connect(func)  # type: ignore


_T = TypeVar("_T")


@deprecated(info="no longer required, and now a no-op")
def without_qt5_compat_wrapper(cls: _T) -> _T:
    return cls



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

"""
See pylib/anki/hooks.py
"""

from __future__ import annotations

# You can find the definitions in ../tools/genhooks_gui.py
from _aqt.hooks import *



"""
System File Locations
Retrieves common system path names on Windows XP/Vista
Depends only on ctypes, and retrieves path locations in Unicode
"""

import ctypes
from ctypes import windll, wintypes  # type: ignore

__license__ = "MIT"
__version__ = "0.2"
__author__ = "Ryan Ginstrom"
__description__ = "Retrieves common Windows system paths as Unicode strings"


class PathConstants:
    """
    Define constants here to avoid dependency on shellcon.
    Put it in a class to avoid polluting namespace
    """

    CSIDL_DESKTOP = 0
    CSIDL_PROGRAMS = 2
    CSIDL_PERSONAL = 5
    CSIDL_FAVORITES = 6
    CSIDL_STARTUP = 7
    CSIDL_RECENT = 8
    CSIDL_SENDTO = 9
    CSIDL_BITBUCKET = 10
    CSIDL_STARTMENU = 11
    CSIDL_MYDOCUMENTS = 12
    CSIDL_MYMUSIC = 13
    CSIDL_MYVIDEO = 14
    CSIDL_DESKTOPDIRECTORY = 16
    CSIDL_DRIVES = 17
    CSIDL_NETWORK = 18
    CSIDL_NETHOOD = 19
    CSIDL_FONTS = 20
    CSIDL_TEMPLATES = 21
    CSIDL_COMMON_STARTMENU = 22
    CSIDL_COMMON_PROGRAMS = 23
    CSIDL_COMMON_STARTUP = 24
    CSIDL_COMMON_DESKTOPDIRECTORY = 25
    CSIDL_APPDATA = 26
    CSIDL_PRINTHOOD = 27
    CSIDL_LOCAL_APPDATA = 28
    CSIDL_ALTSTARTUP = 29
    CSIDL_COMMON_ALTSTARTUP = 30
    CSIDL_COMMON_FAVORITES = 31
    CSIDL_INTERNET_CACHE = 32
    CSIDL_COOKIES = 33
    CSIDL_HISTORY = 34
    CSIDL_COMMON_APPDATA = 35
    CSIDL_WINDOWS = 36
    CSIDL_SYSTEM = 37
    CSIDL_PROGRAM_FILES = 38
    CSIDL_MYPICTURES = 39
    CSIDL_PROFILE = 40
    CSIDL_SYSTEMX86 = 41
    CSIDL_PROGRAM_FILESX86 = 42
    CSIDL_PROGRAM_FILES_COMMON = 43
    CSIDL_PROGRAM_FILES_COMMONX86 = 44
    CSIDL_COMMON_TEMPLATES = 45
    CSIDL_COMMON_DOCUMENTS = 46
    CSIDL_COMMON_ADMINTOOLS = 47
    CSIDL_ADMINTOOLS = 48
    CSIDL_CONNECTIONS = 49
    CSIDL_COMMON_MUSIC = 53
    CSIDL_COMMON_PICTURES = 54
    CSIDL_COMMON_VIDEO = 55
    CSIDL_RESOURCES = 56
    CSIDL_RESOURCES_LOCALIZED = 57
    CSIDL_COMMON_OEM_LINKS = 58
    CSIDL_CDBURN_AREA = 59
    # 60 unused
    CSIDL_COMPUTERSNEARME = 61


class WinPathsException(Exception):
    pass


def _err_unless_zero(result):
    if result == 0:
        return result
    else:
        raise WinPathsException(f"Failed to retrieve windows path: {result}")


_SHGetFolderPath = windll.shell32.SHGetFolderPathW
_SHGetFolderPath.argtypes = [
    wintypes.HWND,
    ctypes.c_int,
    wintypes.HANDLE,
    wintypes.DWORD,
    wintypes.LPCWSTR,
]
_SHGetFolderPath.restype = _err_unless_zero


def _get_path_buf(csidl):
    path_buf = ctypes.create_unicode_buffer(wintypes.MAX_PATH)
    _SHGetFolderPath(0, csidl, 0, 0, path_buf)
    return path_buf.value


def get_local_appdata():
    return _get_path_buf(PathConstants.CSIDL_LOCAL_APPDATA)


def get_appdata():
    return _get_path_buf(PathConstants.CSIDL_APPDATA)


def get_desktop():
    return _get_path_buf(PathConstants.CSIDL_DESKTOP)


def get_programs():
    """current user -> Start menu -> Programs"""
    return _get_path_buf(PathConstants.CSIDL_PROGRAMS)


def get_admin_tools():
    """current user -> Start menu -> Programs -> Admin tools"""
    return _get_path_buf(PathConstants.CSIDL_ADMINTOOLS)


def get_common_admin_tools():
    """all users -> Start menu -> Programs -> Admin tools"""
    return _get_path_buf(PathConstants.CSIDL_COMMON_ADMINTOOLS)


def get_common_appdata():
    return _get_path_buf(PathConstants.CSIDL_COMMON_APPDATA)


def get_common_documents():
    return _get_path_buf(PathConstants.CSIDL_COMMON_DOCUMENTS)


def get_cookies():
    return _get_path_buf(PathConstants.CSIDL_COOKIES)


def get_history():
    return _get_path_buf(PathConstants.CSIDL_HISTORY)


def get_internet_cache():
    return _get_path_buf(PathConstants.CSIDL_INTERNET_CACHE)


def get_my_pictures():
    """Get the user's My Pictures folder"""
    return _get_path_buf(PathConstants.CSIDL_MYPICTURES)


def get_personal():
    """AKA 'My Documents'"""
    return _get_path_buf(PathConstants.CSIDL_PERSONAL)


get_my_documents = get_personal


def get_program_files():
    return _get_path_buf(PathConstants.CSIDL_PROGRAM_FILES)


def get_program_files_common():
    return _get_path_buf(PathConstants.CSIDL_PROGRAM_FILES_COMMON)


def get_system():
    """Use with care and discretion"""
    return _get_path_buf(PathConstants.CSIDL_SYSTEM)


def get_windows():
    """Use with care and discretion"""
    return _get_path_buf(PathConstants.CSIDL_WINDOWS)


def get_favorites():
    return _get_path_buf(PathConstants.CSIDL_FAVORITES)


def get_startup():
    """current user -> start menu -> programs -> startup"""
    return _get_path_buf(PathConstants.CSIDL_STARTUP)


def get_recent():
    return _get_path_buf(PathConstants.CSIDL_RECENT)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html


"""
The V3/2021 scheduler.

https://faqs.ankiweb.net/the-2021-scheduler.html

It uses the same DB schema as the V2 scheduler, and 'schedVer' remains
as '2' internally.
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import Any, Literal

from anki import frontend_pb2, scheduler_pb2
from anki._legacy import deprecated
from anki.cards import Card
from anki.collection import OpChanges
from anki.consts import *
from anki.decks import DeckId
from anki.errors import DBError
from anki.scheduler.legacy import SchedulerBaseWithLegacy
from anki.types import assert_exhaustive
from anki.utils import int_time

QueuedCards = scheduler_pb2.QueuedCards
SchedulingState = scheduler_pb2.SchedulingState
SchedulingStates = scheduler_pb2.SchedulingStates
SchedulingContext = scheduler_pb2.SchedulingContext
SchedulingStatesWithContext = frontend_pb2.SchedulingStatesWithContext
SetSchedulingStatesRequest = frontend_pb2.SetSchedulingStatesRequest
CardAnswer = scheduler_pb2.CardAnswer


class Scheduler(SchedulerBaseWithLegacy):
    version = 3

    # don't rely on this, it will likely be removed in the future
    reps = 0

    # Fetching the next card
    ##########################################################################

    def get_queued_cards(
        self,
        *,
        fetch_limit: int = 1,
        intraday_learning_only: bool = False,
    ) -> QueuedCards:
        "Returns zero or more pending cards, and the remaining counts. Idempotent."
        return self.col._backend.get_queued_cards(
            fetch_limit=fetch_limit, intraday_learning_only=intraday_learning_only
        )

    def describe_next_states(self, next_states: SchedulingStates) -> Sequence[str]:
        "Labels for each of the answer buttons."
        return self.col._backend.describe_next_states(next_states)

    # Answering a card
    ##########################################################################

    def build_answer(
        self,
        *,
        card: Card,
        states: SchedulingStates,
        rating: CardAnswer.Rating.V,
    ) -> CardAnswer:
        "Build input for answer_card()."
        if rating == CardAnswer.AGAIN:
            new_state = states.again
        elif rating == CardAnswer.HARD:
            new_state = states.hard
        elif rating == CardAnswer.GOOD:
            new_state = states.good
        elif rating == CardAnswer.EASY:
            new_state = states.easy
        else:
            raise Exception("invalid rating")

        return CardAnswer(
            card_id=card.id,
            current_state=states.current,
            new_state=new_state,
            rating=rating,
            answered_at_millis=int_time(1000),
            milliseconds_taken=card.time_taken(capped=False),
        )

    def answer_card(self, input: CardAnswer) -> OpChanges:
        "Update card to provided state, and remove it from queue."
        self.reps += 1
        op_bytes = self.col._backend.answer_card_raw(input.SerializeToString())
        return OpChanges.FromString(op_bytes)

    def state_is_leech(self, new_state: SchedulingState) -> bool:
        "True if new state marks the card as a leech."
        return self.col._backend.state_is_leech(new_state)

    # Fetching the next card (legacy API)
    ##########################################################################

    @deprecated(info="no longer required")
    def reset(self) -> None:
        # backend automatically resets queues as operations are performed
        pass

    def getCard(self) -> Card | None:
        """Fetch the next card from the queue. None if finished."""
        try:
            queued_card = self.get_queued_cards().cards[0]
        except IndexError:
            return None

        card = Card(self.col)
        card._load_from_backend_card(queued_card.card)
        card.start_timer()
        return card

    def _is_finished(self) -> bool:
        "Don't use this, it is a stop-gap until this code is refactored."
        return not self.get_queued_cards().cards

    def counts(self, card: Card | None = None) -> tuple[int, int, int]:
        info = self.get_queued_cards()
        return (info.new_count, info.learning_count, info.review_count)

    @property
    def newCount(self) -> int:
        return self.counts()[0]

    @property
    def lrnCount(self) -> int:
        return self.counts()[1]

    @property
    def reviewCount(self) -> int:
        return self.counts()[2]

    def nextIvlStr(self, card: Card, ease: int, short: bool = False) -> str:
        "Return the next interval for CARD as a string."
        states = self.col._backend.get_scheduling_states(card.id)
        return self.col._backend.describe_next_states(states)[ease - 1]

    # Answering a card (legacy API)
    ##########################################################################

    def answerCard(self, card: Card, ease: Literal[1, 2, 3, 4]) -> OpChanges:
        if ease == BUTTON_ONE:
            rating = CardAnswer.AGAIN
        elif ease == BUTTON_TWO:
            rating = CardAnswer.HARD
        elif ease == BUTTON_THREE:
            rating = CardAnswer.GOOD
        elif ease == BUTTON_FOUR:
            rating = CardAnswer.EASY
        else:
            raise Exception("invalid ease")

        states = self.col._backend.get_scheduling_states(card.id)
        changes = self.answer_card(
            self.build_answer(card=card, states=states, rating=rating)
        )

        # tests assume card will be mutated, so we need to reload it
        card.load()

        return changes

    # Next times (legacy API)
    ##########################################################################
    # fixme: move these into tests_schedv2 in the future

    def _interval_for_state(self, state: scheduler_pb2.SchedulingState) -> int:
        kind = state.WhichOneof("kind")
        if kind == "normal":
            return self._interval_for_normal_state(state.normal)
        elif kind == "filtered":
            return self._interval_for_filtered_state(state.filtered)
        else:
            assert_exhaustive(kind)
            return 0

    def _interval_for_normal_state(
        self, normal: scheduler_pb2.SchedulingState.Normal
    ) -> int:
        kind = normal.WhichOneof("kind")
        if kind == "new":
            return 0
        elif kind == "review":
            return normal.review.scheduled_days * 86400
        elif kind == "learning":
            return normal.learning.scheduled_secs
        elif kind == "relearning":
            return normal.relearning.learning.scheduled_secs
        else:
            assert_exhaustive(kind)
            return 0

    def _interval_for_filtered_state(
        self, filtered: scheduler_pb2.SchedulingState.Filtered
    ) -> int:
        kind = filtered.WhichOneof("kind")
        if kind == "preview":
            return filtered.preview.scheduled_secs
        elif kind == "rescheduling":
            return self._interval_for_normal_state(filtered.rescheduling.original_state)
        else:
            assert_exhaustive(kind)
            return 0

    def nextIvl(self, card: Card, ease: int) -> Any:
        "Don't use this - it is only required by tests, and will be moved in the future."
        states = self.col._backend.get_scheduling_states(card.id)
        if ease == BUTTON_ONE:
            new_state = states.again
        elif ease == BUTTON_TWO:
            new_state = states.hard
        elif ease == BUTTON_THREE:
            new_state = states.good
        elif ease == BUTTON_FOUR:
            new_state = states.easy
        else:
            raise Exception("invalid ease")

        return self._interval_for_state(new_state)

    # Other legacy
    ###################

    # called by col.decks.active(), which add-ons are using
    @property
    def active_decks(self) -> list[DeckId]:
        try:
            return self.col.db.list("select id from active_decks")
        except DBError:
            return []



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from collections.abc import Callable, Sequence
from typing import Any, Type, Union

import anki
from anki.collection import Collection
from anki.importing.anki2 import Anki2Importer
from anki.importing.apkg import AnkiPackageImporter
from anki.importing.base import Importer
from anki.importing.csvfile import TextImporter
from anki.importing.mnemo import MnemosyneImporter
from anki.lang import TR


def importers(col: Collection) -> Sequence[tuple[str, type[Importer]]]:
    importers = [
        (col.tr.importing_text_separated_by_tabs_or_semicolons(), TextImporter),
        (
            col.tr.importing_packaged_anki_deckcollection_apkg_colpkg_zip(),
            AnkiPackageImporter,
        ),
        (col.tr.importing_mnemosyne_20_deck_db(), MnemosyneImporter),
    ]
    anki.hooks.importing_importers(importers)
    return importers



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import time
from collections.abc import Callable
from concurrent.futures import Future
from datetime import datetime
from typing import Any

import aqt
import aqt.forms
import aqt.main
from anki.collection import Collection
from anki.errors import Interrupted
from anki.utils import int_time
from aqt import gui_hooks
from aqt.operations import QueryOp
from aqt.qt import QDialog, QDialogButtonBox, QPushButton, Qt, QTimer, qconnect
from aqt.utils import disable_help_button, show_info, tr


class MediaSyncer:
    def __init__(self, mw: aqt.main.AnkiQt) -> None:
        self.mw = mw
        self._syncing: bool = False
        self.last_progress = ""
        self._last_progress_at = 0
        gui_hooks.media_sync_did_start_or_stop.append(self._on_start_stop)

    def start(self, is_periodic_sync: bool = False) -> None:
        "Start media syncing in the background, if it's not already running."
        if not self.mw.pm.media_syncing_enabled() or not (
            auth := self.mw.pm.sync_auth()
        ):
            return

        def run(col: Collection) -> None:
            col.sync_media(auth)

        # this will exit after the thread is spawned, but may block if there's an existing
        # backend lock
        QueryOp(parent=aqt.mw, op=run, success=lambda _: 1).failure(
            lambda e: self._handle_sync_error(e, is_periodic_sync)
        ).run_in_background()

        self.start_monitoring(is_periodic_sync)

    def start_monitoring(self, is_periodic_sync: bool = False) -> None:
        if self._syncing:
            return
        self._syncing = True
        gui_hooks.media_sync_did_start_or_stop(True)
        self._update_progress(tr.sync_media_starting())

        def monitor() -> None:
            while True:
                resp = self.mw.col.media_sync_status()
                if not resp.active:
                    return
                if p := resp.progress:
                    self._update_progress(f"{p.added}, {p.removed}, {p.checked}")

                time.sleep(0.25)

        self.mw.taskman.run_in_background(
            monitor,
            lambda fut: self._on_finished(fut, is_periodic_sync),
            uses_collection=False,
        )

    def _update_progress(self, progress: str) -> None:
        self.last_progress = progress
        self.mw.taskman.run_on_main(lambda: gui_hooks.media_sync_did_progress(progress))

    def _on_finished(self, future: Future, is_periodic_sync: bool = False) -> None:
        self._syncing = False
        self._last_progress_at = int_time()
        gui_hooks.media_sync_did_start_or_stop(False)

        exc = future.exception()
        if exc is not None:
            self._handle_sync_error(exc, is_periodic_sync)
        else:
            self._update_progress(tr.sync_media_complete())

    def _handle_sync_error(
        self, exc: BaseException, is_periodic_sync: bool = False
    ) -> None:
        if isinstance(exc, Interrupted):
            self._update_progress(tr.sync_media_aborted())
        elif is_periodic_sync:
            print(str(exc))
        else:
            show_info(str(exc), modality=Qt.WindowModality.NonModal)

    def abort(self) -> None:
        if not self.is_syncing():
            return
        self.mw.col.set_wants_abort()
        self.mw.col.abort_media_sync()
        self._update_progress(tr.sync_media_aborting())

    def is_syncing(self) -> bool:
        return self._syncing

    def _on_start_stop(self, running: bool) -> None:
        self.mw.toolbar.set_sync_active(running)

    def show_sync_log(self) -> None:
        aqt.dialogs.open("sync_log", self.mw, self)

    def show_diag_until_finished(self, on_finished: Callable[[], None]) -> None:
        # nothing to do if not syncing
        if not self.is_syncing():
            return on_finished()

        diag: MediaSyncDialog = aqt.dialogs.open("sync_log", self.mw, self, True)
        diag.show()

        timer: QTimer

        def check_finished() -> None:
            if not self.is_syncing():
                timer.deleteLater()
                on_finished()

        timer = self.mw.progress.timer(150, check_finished, True, False, parent=self.mw)

    def seconds_since_last_sync(self) -> int:
        if self.is_syncing():
            return 0

        return int_time() - self._last_progress_at


class MediaSyncDialog(QDialog):
    silentlyClose = True

    def __init__(
        self, mw: aqt.main.AnkiQt, syncer: MediaSyncer, close_when_done: bool = False
    ) -> None:
        super().__init__(mw)
        self.mw = mw
        self._syncer = syncer
        self._close_when_done = close_when_done
        self.form = aqt.forms.synclog.Ui_Dialog()
        self.form.setupUi(self)
        self.setWindowTitle(tr.sync_media_log_title())
        disable_help_button(self)
        self.abort_button = QPushButton(tr.sync_abort_button())
        qconnect(self.abort_button.clicked, self._on_abort)
        self.abort_button.setAutoDefault(False)
        self.form.buttonBox.addButton(
            self.abort_button, QDialogButtonBox.ButtonRole.ActionRole
        )
        self.abort_button.setHidden(not self._syncer.is_syncing())

        gui_hooks.media_sync_did_progress.append(self._on_log_entry)
        gui_hooks.media_sync_did_start_or_stop.append(self._on_start_stop)

        self._on_log_entry(syncer.last_progress)
        self.show()

    def reject(self) -> None:
        if self._close_when_done and self._syncer.is_syncing():
            # closing while syncing on close starts an abort
            self._on_abort()
            return

        aqt.dialogs.markClosed("sync_log")
        QDialog.reject(self)

    def reopen(
        self, mw: aqt.AnkiQt, syncer: Any, close_when_done: bool = False
    ) -> None:
        self._close_when_done = close_when_done
        self.show()

    def _on_abort(self, *_args: Any) -> None:
        self._syncer.abort()
        self.abort_button.setHidden(True)

    def _on_log_entry(self, entry: str) -> None:
        dt = datetime.fromtimestamp(int_time())
        time = dt.strftime("%H:%M:%S")
        text = f"{time}: {entry}"
        self.form.log_label.setText(text)
        if not self._syncer.is_syncing():
            self.abort_button.setHidden(True)

    def _on_start_stop(self, running: bool) -> None:
        if not running and self._close_when_done:
            aqt.dialogs.markClosed("sync_log")
            self._close_when_done = False
            self.close()



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

"""
This file contains the Python portion of the template rendering code.

Templates can have filters applied to field replacements. The Rust template
rendering code will apply any built in filters, and stop at the first
unrecognized filter. The remaining filters are returned to Python,
and applied using the hook system. For example,
{{myfilter:hint:text:Field}} will apply the built in text and hint filters,
and then attempt to apply myfilter. If no add-ons have provided the filter,
the filter is skipped.

Add-ons can register a filter with the following code:

from anki import hooks
hooks.field_filter.append(myfunc)

This will call myfunc, passing the field text in as the first argument.
Your function should decide if it wants to modify the text by checking
the filter_name argument, and then return the text whether it has been
modified or not.

A Python implementation of the standard filters is currently available in the
template_legacy.py file, using the legacy addHook() system.
"""

from __future__ import annotations

import os.path
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, Union

import anki
import anki.cards
import anki.collection
import anki.notes
from anki import card_rendering_pb2, hooks
from anki.decks import DeckManager
from anki.errors import TemplateError
from anki.models import NotetypeDict
from anki.sound import AVTag, SoundOrVideoTag, TTSTag
from anki.utils import to_json_bytes


@dataclass
class TemplateReplacement:
    field_name: str
    current_text: str
    filters: list[str]


TemplateReplacementList = list[Union[str, TemplateReplacement]]


@dataclass
class PartiallyRenderedCard:
    qnodes: TemplateReplacementList
    anodes: TemplateReplacementList
    css: str
    latex_svg: bool
    is_empty: bool

    @classmethod
    def from_proto(
        cls, out: card_rendering_pb2.RenderCardResponse
    ) -> PartiallyRenderedCard:
        qnodes = cls.nodes_from_proto(out.question_nodes)
        anodes = cls.nodes_from_proto(out.answer_nodes)

        return PartiallyRenderedCard(
            qnodes, anodes, out.css, out.latex_svg, out.is_empty
        )

    @staticmethod
    def nodes_from_proto(
        nodes: Sequence[card_rendering_pb2.RenderedTemplateNode],
    ) -> TemplateReplacementList:
        results: TemplateReplacementList = []
        for node in nodes:
            if node.WhichOneof("value") == "text":
                results.append(node.text)
            else:
                results.append(
                    TemplateReplacement(
                        field_name=node.replacement.field_name,
                        current_text=node.replacement.current_text,
                        filters=list(node.replacement.filters),
                    )
                )
        return results


def av_tag_to_native(tag: card_rendering_pb2.AVTag) -> AVTag:
    val = tag.WhichOneof("value")
    if val == "sound_or_video":
        return SoundOrVideoTag(filename=os.path.basename(tag.sound_or_video))
    else:
        return TTSTag(
            field_text=tag.tts.field_text,
            lang=tag.tts.lang,
            voices=list(tag.tts.voices),
            other_args=list(tag.tts.other_args),
            speed=tag.tts.speed,
        )


def av_tags_to_native(tags: Sequence[card_rendering_pb2.AVTag]) -> list[AVTag]:
    return list(map(av_tag_to_native, tags))


class TemplateRenderContext:
    """Holds information for the duration of one card render.

    This may fetch information lazily in the future, so please avoid
    using the _private fields directly."""

    @staticmethod
    def from_existing_card(
        card: anki.cards.Card, browser: bool
    ) -> TemplateRenderContext:
        return TemplateRenderContext(card.col, card, card.note(), browser)

    @classmethod
    def from_card_layout(
        cls,
        note: anki.notes.Note,
        card: anki.cards.Card,
        notetype: NotetypeDict,
        template: dict,
        fill_empty: bool,
    ) -> TemplateRenderContext:
        return TemplateRenderContext(
            note.col,
            card,
            note,
            notetype=notetype,
            template=template,
            fill_empty=fill_empty,
        )

    def __init__(
        self,
        col: anki.collection.Collection,
        card: anki.cards.Card,
        note: anki.notes.Note,
        browser: bool = False,
        notetype: NotetypeDict | None = None,
        template: dict | None = None,
        fill_empty: bool = False,
    ) -> None:
        self._col = col.weakref()
        self._card = card
        self._note = note
        self._browser = browser
        self._template = template
        self._fill_empty = fill_empty
        self._fields: dict | None = None
        self._latex_svg = False
        self._question_side: bool = True
        if not notetype:
            self._note_type = note.note_type()
        else:
            self._note_type = notetype

        # if you need to store extra state to share amongst rendering
        # hooks, you can insert it into this dictionary
        self.extra_state: dict[str, Any] = {}

    @property
    def question_side(self) -> bool:
        return self._question_side

    def col(self) -> anki.collection.Collection:
        return self._col

    def fields(self) -> dict[str, str]:
        print(".fields() is obsolete, use .note() or .card()")
        if not self._fields:
            # fields from note
            fields = dict(self._note.items())

            # add (most) special fields
            fields["Tags"] = self._note.string_tags().strip()
            fields["Type"] = self._note_type["name"]
            fields["Deck"] = self._col.decks.name(self._card.current_deck_id())
            fields["Subdeck"] = DeckManager.basename(fields["Deck"])
            if self._template:
                fields["Card"] = self._template["name"]
            else:
                fields["Card"] = ""
            flag = self._card.user_flag()
            fields["CardFlag"] = flag and f"flag{flag}" or ""
            self._fields = fields

        return self._fields

    def card(self) -> anki.cards.Card:
        """Returns the card being rendered.

        Be careful not to call .question() or .answer() on the card, or you'll create an
        infinite loop."""
        return self._card

    def note(self) -> anki.notes.Note:
        return self._note

    def note_type(self) -> NotetypeDict:
        return self._note_type

    def latex_svg(self) -> bool:
        return self._latex_svg

    # legacy
    def qfmt(self) -> str:
        return templates_for_card(self.card(), self._browser)[0]

    # legacy
    def afmt(self) -> str:
        return templates_for_card(self.card(), self._browser)[1]

    def render(self) -> TemplateRenderOutput:
        try:
            partial = self._partially_render()
        except TemplateError as error:
            return TemplateRenderOutput(
                question_text=str(error),
                answer_text=str(error),
                question_av_tags=[],
                answer_av_tags=[],
            )

        self._question_side = True
        qtext = apply_custom_filters(partial.qnodes, self, front_side=None)
        qout = self.col()._backend.extract_av_tags(text=qtext, question_side=True)

        self._question_side = False
        atext = apply_custom_filters(partial.anodes, self, front_side=qout.text)
        aout = self.col()._backend.extract_av_tags(text=atext, question_side=False)

        output = TemplateRenderOutput(
            question_text=qout.text,
            answer_text=aout.text,
            question_av_tags=av_tags_to_native(qout.av_tags),
            answer_av_tags=av_tags_to_native(aout.av_tags),
            css=partial.css,
        )

        self._latex_svg = partial.latex_svg

        if not self._browser:
            hooks.card_did_render(output, self)

        return output

    def _partially_render(self) -> PartiallyRenderedCard:
        if self._template:
            # card layout screen
            out = self._col._backend.render_uncommitted_card_legacy(
                note=self._note._to_backend_note(),
                card_ord=self._card.ord,
                template=to_json_bytes(self._template),
                fill_empty=self._fill_empty,
                partial_render=True,
            )
            # when rendering card layout, the css changes have not been
            # committed; we need the current notetype instance instead
            out.css = self._note_type["css"]
        else:
            # existing card (eg study mode)
            out = self._col._backend.render_existing_card(
                card_id=self._card.id, browser=self._browser, partial_render=True
            )
        return PartiallyRenderedCard.from_proto(out)


@dataclass
class TemplateRenderOutput:
    "Stores the rendered templates and extracted AV tags."

    question_text: str
    answer_text: str
    question_av_tags: list[AVTag]
    answer_av_tags: list[AVTag]
    css: str = ""

    def question_and_style(self) -> str:
        return f"<style>{self.css}</style>{self.question_text}"

    def answer_and_style(self) -> str:
        return f"<style>{self.css}</style>{self.answer_text}"


# legacy
def templates_for_card(card: anki.cards.Card, browser: bool) -> tuple[str, str]:
    template = card.template()
    if browser:
        question, answer = template.get("bqfmt"), template.get("bafmt")
    else:
        question, answer = None, None
    question = question or template.get("qfmt")
    answer = answer or template.get("afmt")
    return question, answer  # type: ignore


def apply_custom_filters(
    rendered: TemplateReplacementList,
    ctx: TemplateRenderContext,
    front_side: str | None,
) -> str:
    "Complete rendering by applying any pending custom filters."
    # template already fully rendered?
    if len(rendered) == 1 and isinstance(rendered[0], str):
        return rendered[0]

    res = ""
    for node in rendered:
        if isinstance(node, str):
            res += node
        else:
            # do we need to inject in FrontSide?
            if node.field_name == "FrontSide" and front_side is not None:
                node.current_text = front_side

            field_text = node.current_text
            for filter_name in node.filters:
                field_text = hooks.field_filter(
                    field_text, node.field_name, filter_name, ctx
                )
                # legacy hook - the second and fifth argument are no longer used.
                field_text = hooks.runFilter(
                    f"fmod_{filter_name}",
                    field_text,
                    "",
                    ctx.note().items(),
                    node.field_name,
                    "",
                )

            res += field_text
    return res



# Form implementation generated from reading ui file 'qt/aqt/forms/forget.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(235, 118)
        Dialog.setModal(True)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.restore_position = QtWidgets.QCheckBox(parent=Dialog)
        self.restore_position.setObjectName("restore_position")
        self.verticalLayout_2.addWidget(self.restore_position, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.reset_counts = QtWidgets.QCheckBox(parent=Dialog)
        self.reset_counts.setObjectName("reset_counts")
        self.verticalLayout_2.addWidget(self.reset_counts, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout_2.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(Dialog.reject) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(tr.actions_forget_card())
        self.restore_position.setText(tr.scheduling_restore_position())
        self.reset_counts.setText(tr.scheduling_reset_counts())


# Form implementation generated from reading ui file 'qt/aqt/forms/widgets.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(925, 822)
        Dialog.setWindowTitle("Qt Widget Gallery")
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.topLayout = QtWidgets.QHBoxLayout()
        self.topLayout.setObjectName("topLayout")
        self.styleLabel = QtWidgets.QLabel(parent=Dialog)
        self.styleLabel.setText("Style")
        self.styleLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight|QtCore.Qt.AlignmentFlag.AlignTrailing|QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.styleLabel.setObjectName("styleLabel")
        self.topLayout.addWidget(self.styleLabel)
        self.styleComboBox = QtWidgets.QComboBox(parent=Dialog)
        self.styleComboBox.setCurrentText("")
        self.styleComboBox.setObjectName("styleComboBox")
        self.topLayout.addWidget(self.styleComboBox)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
        self.topLayout.addItem(spacerItem)
        self.disableCheckBox = QtWidgets.QCheckBox(parent=Dialog)
        self.disableCheckBox.setText("Disable Widgets")
        self.disableCheckBox.setObjectName("disableCheckBox")
        self.topLayout.addWidget(self.disableCheckBox)
        self.verticalLayout.addLayout(self.topLayout)
        self.testGrid = QtWidgets.QWidget(parent=Dialog)
        self.testGrid.setObjectName("testGrid")
        self.gridLayout = QtWidgets.QGridLayout(self.testGrid)
        self.gridLayout.setObjectName("gridLayout")
        self.checkButtonsGroup = QtWidgets.QGroupBox(parent=self.testGrid)
        self.checkButtonsGroup.setTitle("Check Buttons")
        self.checkButtonsGroup.setObjectName("checkButtonsGroup")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.checkButtonsGroup)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.radioButtonNotCheckable = QtWidgets.QRadioButton(parent=self.checkButtonsGroup)
        self.radioButtonNotCheckable.setText("RadioButton (not checkable)")
        self.radioButtonNotCheckable.setCheckable(False)
        self.radioButtonNotCheckable.setObjectName("radioButtonNotCheckable")
        self.verticalLayout_3.addWidget(self.radioButtonNotCheckable)
        self.radioButtonChecked = QtWidgets.QRadioButton(parent=self.checkButtonsGroup)
        self.radioButtonChecked.setText("RadioButton (checked)")
        self.radioButtonChecked.setChecked(True)
        self.radioButtonChecked.setObjectName("radioButtonChecked")
        self.verticalLayout_3.addWidget(self.radioButtonChecked)
        self.radioButtonUnchecked = QtWidgets.QRadioButton(parent=self.checkButtonsGroup)
        self.radioButtonUnchecked.setText("RadioButton (unchecked)")
        self.radioButtonUnchecked.setObjectName("radioButtonUnchecked")
        self.verticalLayout_3.addWidget(self.radioButtonUnchecked)
        self.checkBoxTristate = QtWidgets.QCheckBox(parent=self.checkButtonsGroup)
        self.checkBoxTristate.setText("CheckBox (tristate)")
        self.checkBoxTristate.setChecked(True)
        self.checkBoxTristate.setTristate(True)
        self.checkBoxTristate.setObjectName("checkBoxTristate")
        self.verticalLayout_3.addWidget(self.checkBoxTristate)
        self.gridLayout.addWidget(self.checkButtonsGroup, 0, 0, 1, 1)
        self.buttonsGroup = QtWidgets.QGroupBox(parent=self.testGrid)
        self.buttonsGroup.setTitle("Buttons")
        self.buttonsGroup.setObjectName("buttonsGroup")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.buttonsGroup)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.pushButton = QtWidgets.QPushButton(parent=self.buttonsGroup)
        self.pushButton.setText("PushButton")
        self.pushButton.setObjectName("pushButton")
        self.verticalLayout_4.addWidget(self.pushButton)
        self.pushButtonCheckable = QtWidgets.QPushButton(parent=self.buttonsGroup)
        self.pushButtonCheckable.setText("PushButton (checkable)")
        self.pushButtonCheckable.setCheckable(True)
        self.pushButtonCheckable.setChecked(True)
        self.pushButtonCheckable.setObjectName("pushButtonCheckable")
        self.verticalLayout_4.addWidget(self.pushButtonCheckable)
        self.pushButtonFlat = QtWidgets.QPushButton(parent=self.buttonsGroup)
        self.pushButtonFlat.setText("PushButton (flat)")
        self.pushButtonFlat.setAutoDefault(True)
        self.pushButtonFlat.setFlat(True)
        self.pushButtonFlat.setObjectName("pushButtonFlat")
        self.verticalLayout_4.addWidget(self.pushButtonFlat)
        self.gridLayout.addWidget(self.buttonsGroup, 0, 1, 1, 1)
        self.calendarGroup = QtWidgets.QGroupBox(parent=self.testGrid)
        self.calendarGroup.setTitle("CalendarWidget")
        self.calendarGroup.setObjectName("calendarGroup")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.calendarGroup)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.calendarWidget = QtWidgets.QCalendarWidget(parent=self.calendarGroup)
        self.calendarWidget.setObjectName("calendarWidget")
        self.gridLayout_3.addWidget(self.calendarWidget, 0, 0, 1, 1)
        self.gridLayout.addWidget(self.calendarGroup, 1, 0, 1, 1)
        self.textInputsGroup = QtWidgets.QGroupBox(parent=self.testGrid)
        self.textInputsGroup.setTitle("Text Inputs")
        self.textInputsGroup.setObjectName("textInputsGroup")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.textInputsGroup)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.comboBox = QtWidgets.QComboBox(parent=self.textInputsGroup)
        self.comboBox.setEditable(True)
        self.comboBox.setCurrentText("ComboBox (editable)")
        self.comboBox.setObjectName("comboBox")
        self.verticalLayout_8.addWidget(self.comboBox)
        self.lineEdit = QtWidgets.QLineEdit(parent=self.textInputsGroup)
        self.lineEdit.setText("")
        self.lineEdit.setPlaceholderText("LineEdit")
        self.lineEdit.setObjectName("lineEdit")
        self.verticalLayout_8.addWidget(self.lineEdit)
        self.splitter = QtWidgets.QSplitter(parent=self.textInputsGroup)
        self.splitter.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.splitter.setObjectName("splitter")
        self.plainTextEdit = QtWidgets.QPlainTextEdit(parent=self.splitter)
        self.plainTextEdit.setPlainText("")
        self.plainTextEdit.setPlaceholderText("PlainTextEdit")
        self.plainTextEdit.setObjectName("plainTextEdit")
        self.textEdit = QtWidgets.QTextEdit(parent=self.splitter)
        self.textEdit.setDocumentTitle("")
        self.textEdit.setPlaceholderText("TextEdit")
        self.textEdit.setObjectName("textEdit")
        self.verticalLayout_8.addWidget(self.splitter)
        self.gridLayout.addWidget(self.textInputsGroup, 1, 1, 1, 1)
        self.otherInputsGroup = QtWidgets.QGroupBox(parent=self.testGrid)
        self.otherInputsGroup.setTitle("Other Inputs")
        self.otherInputsGroup.setObjectName("otherInputsGroup")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.otherInputsGroup)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.keySequenceEdit = QtWidgets.QKeySequenceEdit(parent=self.otherInputsGroup)
        self.keySequenceEdit.setKeySequence("")
        self.keySequenceEdit.setObjectName("keySequenceEdit")
        self.gridLayout_2.addWidget(self.keySequenceEdit, 2, 1, 1, 2)
        self.spinBox = QtWidgets.QSpinBox(parent=self.otherInputsGroup)
        self.spinBox.setSuffix("")
        self.spinBox.setPrefix("")
        self.spinBox.setProperty("value", 1)
        self.spinBox.setObjectName("spinBox")
        self.gridLayout_2.addWidget(self.spinBox, 0, 1, 1, 2)
        self.keySequenceLabel = QtWidgets.QLabel(parent=self.otherInputsGroup)
        self.keySequenceLabel.setText("KeySequenceEdit")
        self.keySequenceLabel.setObjectName("keySequenceLabel")
        self.gridLayout_2.addWidget(self.keySequenceLabel, 2, 0, 1, 1)
        self.dateTimeLabel = QtWidgets.QLabel(parent=self.otherInputsGroup)
        self.dateTimeLabel.setText("DateTimeEdit")
        self.dateTimeLabel.setObjectName("dateTimeLabel")
        self.gridLayout_2.addWidget(self.dateTimeLabel, 1, 0, 1, 1)
        self.spinBoxLabel = QtWidgets.QLabel(parent=self.otherInputsGroup)
        self.spinBoxLabel.setText("SpinBox")
        self.spinBoxLabel.setObjectName("spinBoxLabel")
        self.gridLayout_2.addWidget(self.spinBoxLabel, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.sliderLabel = QtWidgets.QLabel(parent=self.otherInputsGroup)
        self.sliderLabel.setText("Slider")
        self.sliderLabel.setObjectName("sliderLabel")
        self.horizontalLayout.addWidget(self.sliderLabel)
        self.horizontalSlider = QtWidgets.QSlider(parent=self.otherInputsGroup)
        self.horizontalSlider.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.horizontalSlider.setObjectName("horizontalSlider")
        self.horizontalLayout.addWidget(self.horizontalSlider)
        self.label = QtWidgets.QLabel(parent=self.otherInputsGroup)
        self.label.setText("Dial")
        self.label.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight|QtCore.Qt.AlignmentFlag.AlignTrailing|QtCore.Qt.AlignmentFlag.AlignVCenter)
        self.label.setObjectName("label")
        self.horizontalLayout.addWidget(self.label)
        self.dial = QtWidgets.QDial(parent=self.otherInputsGroup)
        self.dial.setObjectName("dial")
        self.horizontalLayout.addWidget(self.dial)
        self.gridLayout_2.addLayout(self.horizontalLayout, 3, 0, 1, 3)
        self.dateTimeEdit = QtWidgets.QDateTimeEdit(parent=self.otherInputsGroup)
        self.dateTimeEdit.setObjectName("dateTimeEdit")
        self.gridLayout_2.addWidget(self.dateTimeEdit, 1, 1, 1, 2)
        self.gridLayout.addWidget(self.otherInputsGroup, 2, 0, 1, 1)
        self.tabWidget = QtWidgets.QTabWidget(parent=self.testGrid)
        self.tabWidget.setObjectName("tabWidget")
        self.listTab = QtWidgets.QWidget()
        self.listTab.setObjectName("listTab")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.listTab)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.listWidget = QtWidgets.QListWidget(parent=self.listTab)
        self.listWidget.setObjectName("listWidget")
        self.verticalLayout_6.addWidget(self.listWidget)
        self.tabWidget.addTab(self.listTab, "ListWidget")
        self.treeTab = QtWidgets.QWidget()
        self.treeTab.setObjectName("treeTab")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.treeTab)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.treeWidget = QtWidgets.QTreeWidget(parent=self.treeTab)
        self.treeWidget.setObjectName("treeWidget")
        self.treeWidget.headerItem().setText(0, "1")
        self.verticalLayout_7.addWidget(self.treeWidget)
        self.tabWidget.addTab(self.treeTab, "TreeWidget")
        self.tableTab = QtWidgets.QWidget()
        self.tableTab.setObjectName("tableTab")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.tableTab)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.tableWidget = QtWidgets.QTableWidget(parent=self.tableTab)
        self.tableWidget.setRowCount(0)
        self.tableWidget.setColumnCount(0)
        self.tableWidget.setObjectName("tableWidget")
        self.verticalLayout_5.addWidget(self.tableWidget)
        self.tabWidget.addTab(self.tableTab, "TableWidget")
        self.gridLayout.addWidget(self.tabWidget, 2, 1, 1, 1)
        self.progressBarLayout = QtWidgets.QHBoxLayout()
        self.progressBarLayout.setObjectName("progressBarLayout")
        self.progressBarLabel = QtWidgets.QLabel(parent=self.testGrid)
        self.progressBarLabel.setText("ProgressBar")
        self.progressBarLabel.setObjectName("progressBarLabel")
        self.progressBarLayout.addWidget(self.progressBarLabel)
        self.progressBar = QtWidgets.QProgressBar(parent=self.testGrid)
        self.progressBar.setProperty("value", 24)
        self.progressBar.setObjectName("progressBar")
        self.progressBarLayout.addWidget(self.progressBar)
        self.gridLayout.addLayout(self.progressBarLayout, 3, 0, 1, 2)
        self.verticalLayout.addWidget(self.testGrid)

        self.retranslateUi(Dialog)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        pass


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/links.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/links.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import generic_pb2 as anki_dot_generic__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x61nki/links.proto\x12\nanki.links\x1a\x12\x61nki/generic.proto\"\xbc\x04\n\x13HelpPageLinkRequest\x12\x36\n\x04page\x18\x01 \x01(\x0e\x32(.anki.links.HelpPageLinkRequest.HelpPage\"\xec\x03\n\x08HelpPage\x12\r\n\tNOTE_TYPE\x10\x00\x12\x0c\n\x08\x42ROWSING\x10\x01\x12\x1d\n\x19\x42ROWSING_FIND_AND_REPLACE\x10\x02\x12\x17\n\x13\x42ROWSING_NOTES_MENU\x10\x03\x12\x16\n\x12KEYBOARD_SHORTCUTS\x10\x04\x12\x0b\n\x07\x45\x44ITING\x10\x05\x12\x18\n\x14\x41\x44\x44ING_CARD_AND_NOTE\x10\x06\x12\x16\n\x12\x41\x44\x44ING_A_NOTE_TYPE\x10\x07\x12\t\n\x05LATEX\x10\x08\x12\x0f\n\x0bPREFERENCES\x10\t\x12\t\n\x05INDEX\x10\n\x12\r\n\tTEMPLATES\x10\x0b\x12\x11\n\rFILTERED_DECK\x10\x0c\x12\r\n\tIMPORTING\x10\r\x12\x16\n\x12\x43USTOMIZING_FIELDS\x10\x0e\x12\x10\n\x0c\x44\x45\x43K_OPTIONS\x10\x0f\x12\x14\n\x10\x45\x44ITING_FEATURES\x10\x10\x12\x15\n\x11\x46ULL_SCREEN_ISSUE\x10\x11\x12\x17\n\x13\x43\x41RD_TYPE_DUPLICATE\x10\x12\x12\x1c\n\x18\x43\x41RD_TYPE_NO_FRONT_FIELD\x10\x13\x12\x1b\n\x17\x43\x41RD_TYPE_MISSING_CLOZE\x10\x14\x12\x13\n\x0fTROUBLESHOOTING\x10\x15\x12\x1c\n\x18\x43\x41RD_TYPE_TEMPLATE_ERROR\x10\x16\x32U\n\x0cLinksService\x12\x45\n\x0cHelpPageLink\x12\x1f.anki.links.HelpPageLinkRequest\x1a\x14.anki.generic.String2\x15\n\x13\x42\x61\x63kendLinksServiceB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.links_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_HELPPAGELINKREQUEST']._serialized_start=53
  _globals['_HELPPAGELINKREQUEST']._serialized_end=625
  _globals['_HELPPAGELINKREQUEST_HELPPAGE']._serialized_start=133
  _globals['_HELPPAGELINKREQUEST_HELPPAGE']._serialized_end=625
  _globals['_LINKSSERVICE']._serialized_start=627
  _globals['_LINKSSERVICE']._serialized_end=712
  _globals['_BACKENDLINKSSERVICE']._serialized_start=714
  _globals['_BACKENDLINKSSERVICE']._serialized_end=735
# @@protoc_insertion_point(module_scope)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
from __future__ import annotations

import aqt
import aqt.browser
from aqt.browser.sidebar.item import SidebarItem
from aqt.qt import *
from aqt.theme import theme_manager


class SidebarModel(QAbstractItemModel):
    def __init__(
        self, sidebar: aqt.browser.sidebar.SidebarTreeView, root: SidebarItem
    ) -> None:
        super().__init__(sidebar)
        self.sidebar = sidebar
        self.root = root
        self._cache_rows(root)

    def _cache_rows(self, node: SidebarItem) -> None:
        "Cache index of children in parent."
        for row, item in enumerate(node.children):
            item._row_in_parent = row
            self._cache_rows(item)

    def item_for_index(self, idx: QModelIndex) -> SidebarItem:
        return idx.internalPointer()

    def index_for_item(self, item: SidebarItem) -> QModelIndex:
        assert item._row_in_parent is not None
        return self.createIndex(item._row_in_parent, 0, item)

    def search(self, text: str) -> bool:
        return self.root.search(text.lower())

    # Qt API
    ######################################################################

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        if not parent.isValid():
            return len(self.root.children)
        else:
            item: SidebarItem = parent.internalPointer()
            return len(item.children)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return 1

    def index(
        self, row: int, column: int, parent: QModelIndex = QModelIndex()
    ) -> QModelIndex:
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        parentItem: SidebarItem
        if not parent.isValid():
            parentItem = self.root
        else:
            parentItem = parent.internalPointer()

        item = parentItem.children[row]
        return self.createIndex(row, column, item)

    def parent(self, child: QModelIndex) -> QModelIndex:  # type: ignore
        if not child.isValid():
            return QModelIndex()

        childItem: SidebarItem = child.internalPointer()
        parentItem = childItem._parent_item

        if parentItem is None or parentItem == self.root:
            return QModelIndex()

        row = parentItem._row_in_parent
        assert row is not None

        return self.createIndex(row, 0, parentItem)

    def data(
        self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole
    ) -> QVariant:
        if not index.isValid():
            return QVariant()

        if role not in (
            Qt.ItemDataRole.DisplayRole,
            Qt.ItemDataRole.DecorationRole,
            Qt.ItemDataRole.ToolTipRole,
            Qt.ItemDataRole.EditRole,
        ):
            return QVariant()

        item: SidebarItem = index.internalPointer()

        if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.EditRole):
            return QVariant(item.name)
        if role == Qt.ItemDataRole.ToolTipRole:
            return QVariant(item.tooltip)
        return QVariant(theme_manager.icon_from_resources(item.icon))

    def setData(
        self, index: QModelIndex, text: str, _role: int = Qt.ItemDataRole.EditRole
    ) -> bool:
        return self.sidebar._on_rename(index.internalPointer(), text)

    def supportedDropActions(self) -> Qt.DropAction:
        return Qt.DropAction.MoveAction

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:
        if not index.isValid():
            return Qt.ItemFlag.ItemIsEnabled
        flags = (
            Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsSelectable
            | Qt.ItemFlag.ItemIsDragEnabled
        )
        item: SidebarItem = index.internalPointer()
        if item.item_type in self.sidebar.valid_drop_types:
            flags |= Qt.ItemFlag.ItemIsDropEnabled
        if item.item_type.is_editable():
            flags |= Qt.ItemFlag.ItemIsEditable

        return flags



# Form implementation generated from reading ui file 'qt/aqt/forms/findreplace.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(479, 247)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.replace = QtWidgets.QComboBox(parent=Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(9)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.replace.sizePolicy().hasHeightForWidth())
        self.replace.setSizePolicy(sizePolicy)
        self.replace.setEditable(True)
        self.replace.setInsertPolicy(QtWidgets.QComboBox.InsertPolicy.NoInsert)
        self.replace.setSizeAdjustPolicy(QtWidgets.QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.replace.setObjectName("replace")
        self.gridLayout.addWidget(self.replace, 1, 1, 1, 1)
        self.label_2 = QtWidgets.QLabel(parent=Dialog)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)
        self.label_3 = QtWidgets.QLabel(parent=Dialog)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 2, 0, 1, 1)
        self.label = QtWidgets.QLabel(parent=Dialog)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.field = QtWidgets.QComboBox(parent=Dialog)
        self.field.setSizeAdjustPolicy(QtWidgets.QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.field.setObjectName("field")
        self.gridLayout.addWidget(self.field, 2, 1, 1, 1)
        self.re = QtWidgets.QCheckBox(parent=Dialog)
        self.re.setObjectName("re")
        self.gridLayout.addWidget(self.re, 5, 1, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.ignoreCase = QtWidgets.QCheckBox(parent=Dialog)
        self.ignoreCase.setChecked(True)
        self.ignoreCase.setObjectName("ignoreCase")
        self.gridLayout.addWidget(self.ignoreCase, 4, 1, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.find = QtWidgets.QComboBox(parent=Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(9)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.find.sizePolicy().hasHeightForWidth())
        self.find.setSizePolicy(sizePolicy)
        self.find.setEditable(True)
        self.find.setInsertPolicy(QtWidgets.QComboBox.InsertPolicy.NoInsert)
        self.find.setSizeAdjustPolicy(QtWidgets.QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.find.setObjectName("find")
        self.gridLayout.addWidget(self.find, 0, 1, 1, 1)
        self.selected_notes = QtWidgets.QCheckBox(parent=Dialog)
        self.selected_notes.setChecked(True)
        self.selected_notes.setObjectName("selected_notes")
        self.gridLayout.addWidget(self.selected_notes, 3, 1, 1, 1, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.verticalLayout.addLayout(self.gridLayout)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Help|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(Dialog.reject) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        Dialog.setTabOrder(self.find, self.replace)
        Dialog.setTabOrder(self.replace, self.field)
        Dialog.setTabOrder(self.field, self.selected_notes)
        Dialog.setTabOrder(self.selected_notes, self.ignoreCase)
        Dialog.setTabOrder(self.ignoreCase, self.re)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(tr.browsing_find_and_replace())
        self.label_2.setText(tr.browsing_replace_with())
        self.label_3.setText(tr.browsing_in())
        self.label.setText(tr.browsing_find())
        self.re.setText(tr.browsing_treat_input_as_regular_expression())
        self.ignoreCase.setText(tr.browsing_ignore_case())
        self.selected_notes.setText(tr.browsing_selected_notes_only())


# Form implementation generated from reading ui file 'qt/aqt/forms/addmodel.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(285, 269)
        self.vboxlayout = QtWidgets.QVBoxLayout(Dialog)
        self.vboxlayout.setObjectName("vboxlayout")
        self.groupBox = QtWidgets.QGroupBox(parent=Dialog)
        self.groupBox.setTitle("")
        self.groupBox.setObjectName("groupBox")
        self.vboxlayout1 = QtWidgets.QVBoxLayout(self.groupBox)
        self.vboxlayout1.setObjectName("vboxlayout1")
        self.models = QtWidgets.QListWidget(parent=self.groupBox)
        self.models.setEditTriggers(QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers)
        self.models.setTabKeyNavigation(True)
        self.models.setObjectName("models")
        self.vboxlayout1.addWidget(self.models)
        self.vboxlayout.addWidget(self.groupBox)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Help|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.vboxlayout.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(Dialog.reject) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(tr.notetypes_add_note_type())


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/import_export.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/import_export.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import cards_pb2 as anki_dot_cards__pb2
from anki import collection_pb2 as anki_dot_collection__pb2
from anki import notes_pb2 as anki_dot_notes__pb2
from anki import generic_pb2 as anki_dot_generic__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x61nki/import_export.proto\x12\x12\x61nki.import_export\x1a\x10\x61nki/cards.proto\x1a\x15\x61nki/collection.proto\x1a\x10\x61nki/notes.proto\x1a\x12\x61nki/generic.proto\"o\n\x1eImportCollectionPackageRequest\x12\x10\n\x08\x63ol_path\x18\x01 \x01(\t\x12\x13\n\x0b\x62\x61\x63kup_path\x18\x02 \x01(\t\x12\x14\n\x0cmedia_folder\x18\x03 \x01(\t\x12\x10\n\x08media_db\x18\x04 \x01(\t\"Y\n\x1e\x45xportCollectionPackageRequest\x12\x10\n\x08out_path\x18\x01 \x01(\t\x12\x15\n\rinclude_media\x18\x02 \x01(\x08\x12\x0e\n\x06legacy\x18\x03 \x01(\x08\"\x83\x02\n\x18ImportAnkiPackageOptions\x12\x17\n\x0fmerge_notetypes\x18\x01 \x01(\x08\x12J\n\x0cupdate_notes\x18\x02 \x01(\x0e\x32\x34.anki.import_export.ImportAnkiPackageUpdateCondition\x12N\n\x10update_notetypes\x18\x03 \x01(\x0e\x32\x34.anki.import_export.ImportAnkiPackageUpdateCondition\x12\x17\n\x0fwith_scheduling\x18\x04 \x01(\x08\x12\x19\n\x11with_deck_configs\x18\x05 \x01(\x08\"o\n\x18ImportAnkiPackageRequest\x12\x14\n\x0cpackage_path\x18\x01 \x01(\t\x12=\n\x07options\x18\x02 \x01(\x0b\x32,.anki.import_export.ImportAnkiPackageOptions\"\x84\x06\n\x0eImportResponse\x12+\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x1a.anki.collection.OpChanges\x12\x33\n\x03log\x18\x02 \x01(\x0b\x32&.anki.import_export.ImportResponse.Log\x1a\x36\n\x04Note\x12\x1e\n\x02id\x18\x01 \x01(\x0b\x32\x12.anki.notes.NoteId\x12\x0e\n\x06\x66ields\x18\x02 \x03(\t\x1a\xd7\x04\n\x03Log\x12\x34\n\x03new\x18\x01 \x03(\x0b\x32\'.anki.import_export.ImportResponse.Note\x12\x38\n\x07updated\x18\x02 \x03(\x0b\x32\'.anki.import_export.ImportResponse.Note\x12:\n\tduplicate\x18\x03 \x03(\x0b\x32\'.anki.import_export.ImportResponse.Note\x12<\n\x0b\x63onflicting\x18\x04 \x03(\x0b\x32\'.anki.import_export.ImportResponse.Note\x12\x42\n\x11\x66irst_field_match\x18\x05 \x03(\x0b\x32\'.anki.import_export.ImportResponse.Note\x12\x41\n\x10missing_notetype\x18\x06 \x03(\x0b\x32\'.anki.import_export.ImportResponse.Note\x12=\n\x0cmissing_deck\x18\x07 \x03(\x0b\x32\'.anki.import_export.ImportResponse.Note\x12\x42\n\x11\x65mpty_first_field\x18\x08 \x03(\x0b\x32\'.anki.import_export.ImportResponse.Note\x12G\n\x0f\x64upe_resolution\x18\t \x01(\x0e\x32..anki.import_export.CsvMetadata.DupeResolution\x12\x13\n\x0b\x66ound_notes\x18\n \x01(\r\"\x9b\x01\n\x18\x45xportAnkiPackageRequest\x12\x10\n\x08out_path\x18\x01 \x01(\t\x12=\n\x07options\x18\x02 \x01(\x0b\x32,.anki.import_export.ExportAnkiPackageOptions\x12.\n\x05limit\x18\x03 \x01(\x0b\x32\x1f.anki.import_export.ExportLimit\"r\n\x18\x45xportAnkiPackageOptions\x12\x17\n\x0fwith_scheduling\x18\x01 \x01(\x08\x12\x19\n\x11with_deck_configs\x18\x02 \x01(\x08\x12\x12\n\nwith_media\x18\x03 \x01(\x08\x12\x0e\n\x06legacy\x18\x04 \x01(\x08\"\xaf\x01\n\x0fPackageMetadata\x12<\n\x07version\x18\x01 \x01(\x0e\x32+.anki.import_export.PackageMetadata.Version\"^\n\x07Version\x12\x13\n\x0fVERSION_UNKNOWN\x10\x00\x12\x14\n\x10VERSION_LEGACY_1\x10\x01\x12\x14\n\x10VERSION_LEGACY_2\x10\x02\x12\x12\n\x0eVERSION_LATEST\x10\x03\"\xbf\x01\n\x0cMediaEntries\x12<\n\x07\x65ntries\x18\x01 \x03(\x0b\x32+.anki.import_export.MediaEntries.MediaEntry\x1aq\n\nMediaEntry\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04size\x18\x02 \x01(\r\x12\x0c\n\x04sha1\x18\x03 \x01(\x0c\x12!\n\x13legacy_zip_filename\x18\xff\x01 \x01(\rH\x00\x88\x01\x01\x42\x16\n\x14_legacy_zip_filename\"S\n\x10ImportCsvRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x31\n\x08metadata\x18\x02 \x01(\x0b\x32\x1f.anki.import_export.CsvMetadata\"\xe1\x01\n\x12\x43svMetadataRequest\x12\x0c\n\x04path\x18\x01 \x01(\t\x12\x41\n\tdelimiter\x18\x02 \x01(\x0e\x32).anki.import_export.CsvMetadata.DelimiterH\x00\x88\x01\x01\x12\x18\n\x0bnotetype_id\x18\x03 \x01(\x03H\x01\x88\x01\x01\x12\x14\n\x07\x64\x65\x63k_id\x18\x04 \x01(\x03H\x02\x88\x01\x01\x12\x14\n\x07is_html\x18\x05 \x01(\x08H\x03\x88\x01\x01\x42\x0c\n\n_delimiterB\x0e\n\x0c_notetype_idB\n\n\x08_deck_idB\n\n\x08_is_html\"\xd9\x06\n\x0b\x43svMetadata\x12<\n\tdelimiter\x18\x01 \x01(\x0e\x32).anki.import_export.CsvMetadata.Delimiter\x12\x0f\n\x07is_html\x18\x02 \x01(\x08\x12\x13\n\x0bglobal_tags\x18\x03 \x03(\t\x12\x14\n\x0cupdated_tags\x18\x04 \x03(\t\x12\x15\n\rcolumn_labels\x18\x05 \x03(\t\x12\x11\n\x07\x64\x65\x63k_id\x18\x06 \x01(\x03H\x00\x12\x15\n\x0b\x64\x65\x63k_column\x18\x07 \x01(\rH\x00\x12\x13\n\tdeck_name\x18\x11 \x01(\tH\x00\x12I\n\x0fglobal_notetype\x18\x08 \x01(\x0b\x32..anki.import_export.CsvMetadata.MappedNotetypeH\x01\x12\x19\n\x0fnotetype_column\x18\t \x01(\rH\x01\x12\x13\n\x0btags_column\x18\n \x01(\r\x12\x17\n\x0f\x66orce_delimiter\x18\x0b \x01(\x08\x12\x15\n\rforce_is_html\x18\x0c \x01(\x08\x12)\n\x07preview\x18\r \x03(\x0b\x32\x18.anki.generic.StringList\x12\x13\n\x0bguid_column\x18\x0e \x01(\r\x12G\n\x0f\x64upe_resolution\x18\x0f \x01(\x0e\x32..anki.import_export.CsvMetadata.DupeResolution\x12?\n\x0bmatch_scope\x18\x10 \x01(\x0e\x32*.anki.import_export.CsvMetadata.MatchScope\x1a\x33\n\x0eMappedNotetype\x12\n\n\x02id\x18\x01 \x01(\x03\x12\x15\n\rfield_columns\x18\x02 \x03(\r\"9\n\x0e\x44upeResolution\x12\n\n\x06UPDATE\x10\x00\x12\x0c\n\x08PRESERVE\x10\x01\x12\r\n\tDUPLICATE\x10\x02\"N\n\tDelimiter\x12\x07\n\x03TAB\x10\x00\x12\x08\n\x04PIPE\x10\x01\x12\r\n\tSEMICOLON\x10\x02\x12\t\n\x05\x43OLON\x10\x03\x12\t\n\x05\x43OMMA\x10\x04\x12\t\n\x05SPACE\x10\x05\"1\n\nMatchScope\x12\x0c\n\x08NOTETYPE\x10\x00\x12\x15\n\x11NOTETYPE_AND_DECK\x10\x01\x42\x06\n\x04\x64\x65\x63kB\n\n\x08notetype\"k\n\x14\x45xportCardCsvRequest\x12\x10\n\x08out_path\x18\x01 \x01(\t\x12\x11\n\twith_html\x18\x02 \x01(\x08\x12.\n\x05limit\x18\x03 \x01(\x0b\x32\x1f.anki.import_export.ExportLimit\"\xbb\x01\n\x14\x45xportNoteCsvRequest\x12\x10\n\x08out_path\x18\x01 \x01(\t\x12\x11\n\twith_html\x18\x02 \x01(\x08\x12\x11\n\twith_tags\x18\x03 \x01(\x08\x12\x11\n\twith_deck\x18\x04 \x01(\x08\x12\x15\n\rwith_notetype\x18\x05 \x01(\x08\x12\x11\n\twith_guid\x18\x06 \x01(\x08\x12.\n\x05limit\x18\x07 \x01(\x0b\x32\x1f.anki.import_export.ExportLimit\"\xac\x01\n\x0b\x45xportLimit\x12/\n\x10whole_collection\x18\x01 \x01(\x0b\x32\x13.anki.generic.EmptyH\x00\x12\x11\n\x07\x64\x65\x63k_id\x18\x02 \x01(\x03H\x00\x12\'\n\x08note_ids\x18\x03 \x01(\x0b\x32\x13.anki.notes.NoteIdsH\x00\x12\'\n\x08\x63\x61rd_ids\x18\x04 \x01(\x0b\x32\x13.anki.cards.CardIdsH\x00\x42\x07\n\x05limit*\xb6\x01\n ImportAnkiPackageUpdateCondition\x12\x31\n-IMPORT_ANKI_PACKAGE_UPDATE_CONDITION_IF_NEWER\x10\x00\x12/\n+IMPORT_ANKI_PACKAGE_UPDATE_CONDITION_ALWAYS\x10\x01\x12.\n*IMPORT_ANKI_PACKAGE_UPDATE_CONDITION_NEVER\x10\x02\x32\xa5\x06\n\x13ImportExportService\x12\x65\n\x11ImportAnkiPackage\x12,.anki.import_export.ImportAnkiPackageRequest\x1a\".anki.import_export.ImportResponse\x12`\n\x1bGetImportAnkiPackagePresets\x12\x13.anki.generic.Empty\x1a,.anki.import_export.ImportAnkiPackageOptions\x12W\n\x11\x45xportAnkiPackage\x12,.anki.import_export.ExportAnkiPackageRequest\x1a\x14.anki.generic.UInt32\x12Y\n\x0eGetCsvMetadata\x12&.anki.import_export.CsvMetadataRequest\x1a\x1f.anki.import_export.CsvMetadata\x12U\n\tImportCsv\x12$.anki.import_export.ImportCsvRequest\x1a\".anki.import_export.ImportResponse\x12O\n\rExportNoteCsv\x12(.anki.import_export.ExportNoteCsvRequest\x1a\x14.anki.generic.UInt32\x12O\n\rExportCardCsv\x12(.anki.import_export.ExportCardCsvRequest\x1a\x14.anki.generic.UInt32\x12J\n\x0eImportJsonFile\x12\x14.anki.generic.String\x1a\".anki.import_export.ImportResponse\x12L\n\x10ImportJsonString\x12\x14.anki.generic.String\x1a\".anki.import_export.ImportResponse2\xe4\x01\n\x1a\x42\x61\x63kendImportExportService\x12\x62\n\x17ImportCollectionPackage\x12\x32.anki.import_export.ImportCollectionPackageRequest\x1a\x13.anki.generic.Empty\x12\x62\n\x17\x45xportCollectionPackage\x12\x32.anki.import_export.ExportCollectionPackageRequest\x1a\x13.anki.generic.EmptyB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.import_export_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_IMPORTANKIPACKAGEUPDATECONDITION']._serialized_start=3775
  _globals['_IMPORTANKIPACKAGEUPDATECONDITION']._serialized_end=3957
  _globals['_IMPORTCOLLECTIONPACKAGEREQUEST']._serialized_start=127
  _globals['_IMPORTCOLLECTIONPACKAGEREQUEST']._serialized_end=238
  _globals['_EXPORTCOLLECTIONPACKAGEREQUEST']._serialized_start=240
  _globals['_EXPORTCOLLECTIONPACKAGEREQUEST']._serialized_end=329
  _globals['_IMPORTANKIPACKAGEOPTIONS']._serialized_start=332
  _globals['_IMPORTANKIPACKAGEOPTIONS']._serialized_end=591
  _globals['_IMPORTANKIPACKAGEREQUEST']._serialized_start=593
  _globals['_IMPORTANKIPACKAGEREQUEST']._serialized_end=704
  _globals['_IMPORTRESPONSE']._serialized_start=707
  _globals['_IMPORTRESPONSE']._serialized_end=1479
  _globals['_IMPORTRESPONSE_NOTE']._serialized_start=823
  _globals['_IMPORTRESPONSE_NOTE']._serialized_end=877
  _globals['_IMPORTRESPONSE_LOG']._serialized_start=880
  _globals['_IMPORTRESPONSE_LOG']._serialized_end=1479
  _globals['_EXPORTANKIPACKAGEREQUEST']._serialized_start=1482
  _globals['_EXPORTANKIPACKAGEREQUEST']._serialized_end=1637
  _globals['_EXPORTANKIPACKAGEOPTIONS']._serialized_start=1639
  _globals['_EXPORTANKIPACKAGEOPTIONS']._serialized_end=1753
  _globals['_PACKAGEMETADATA']._serialized_start=1756
  _globals['_PACKAGEMETADATA']._serialized_end=1931
  _globals['_PACKAGEMETADATA_VERSION']._serialized_start=1837
  _globals['_PACKAGEMETADATA_VERSION']._serialized_end=1931
  _globals['_MEDIAENTRIES']._serialized_start=1934
  _globals['_MEDIAENTRIES']._serialized_end=2125
  _globals['_MEDIAENTRIES_MEDIAENTRY']._serialized_start=2012
  _globals['_MEDIAENTRIES_MEDIAENTRY']._serialized_end=2125
  _globals['_IMPORTCSVREQUEST']._serialized_start=2127
  _globals['_IMPORTCSVREQUEST']._serialized_end=2210
  _globals['_CSVMETADATAREQUEST']._serialized_start=2213
  _globals['_CSVMETADATAREQUEST']._serialized_end=2438
  _globals['_CSVMETADATA']._serialized_start=2441
  _globals['_CSVMETADATA']._serialized_end=3298
  _globals['_CSVMETADATA_MAPPEDNOTETYPE']._serialized_start=3037
  _globals['_CSVMETADATA_MAPPEDNOTETYPE']._serialized_end=3088
  _globals['_CSVMETADATA_DUPERESOLUTION']._serialized_start=3090
  _globals['_CSVMETADATA_DUPERESOLUTION']._serialized_end=3147
  _globals['_CSVMETADATA_DELIMITER']._serialized_start=3149
  _globals['_CSVMETADATA_DELIMITER']._serialized_end=3227
  _globals['_CSVMETADATA_MATCHSCOPE']._serialized_start=3229
  _globals['_CSVMETADATA_MATCHSCOPE']._serialized_end=3278
  _globals['_EXPORTCARDCSVREQUEST']._serialized_start=3300
  _globals['_EXPORTCARDCSVREQUEST']._serialized_end=3407
  _globals['_EXPORTNOTECSVREQUEST']._serialized_start=3410
  _globals['_EXPORTNOTECSVREQUEST']._serialized_end=3597
  _globals['_EXPORTLIMIT']._serialized_start=3600
  _globals['_EXPORTLIMIT']._serialized_end=3772
  _globals['_IMPORTEXPORTSERVICE']._serialized_start=3960
  _globals['_IMPORTEXPORTSERVICE']._serialized_end=4765
  _globals['_BACKENDIMPORTEXPORTSERVICE']._serialized_start=4768
  _globals['_BACKENDIMPORTEXPORTSERVICE']._serialized_end=4996
# @@protoc_insertion_point(module_scope)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import copy
from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING, Any, NewType

if TYPE_CHECKING:
    import anki

import anki.cards
import anki.collection
from anki import deck_config_pb2, decks_pb2
from anki._legacy import DeprecatedNamesMixin, deprecated, print_deprecation_warning
from anki.collection import OpChanges, OpChangesWithCount, OpChangesWithId
from anki.consts import *
from anki.errors import NotFoundError
from anki.utils import from_json_bytes, ids2str, int_time, to_json_bytes

# public exports
DeckTreeNode = decks_pb2.DeckTreeNode
DeckNameId = decks_pb2.DeckNameId
FilteredDeckConfig = decks_pb2.Deck.Filtered
DeckCollapseScope = decks_pb2.SetDeckCollapsedRequest.Scope
DeckConfigsForUpdate = deck_config_pb2.DeckConfigsForUpdate
UpdateDeckConfigs = deck_config_pb2.UpdateDeckConfigsRequest
Deck = decks_pb2.Deck

# type aliases until we can move away from dicts
DeckDict = dict[str, Any]
DeckConfigDict = dict[str, Any]

DeckId = NewType("DeckId", int)
DeckConfigId = NewType("DeckConfigId", int)

DEFAULT_DECK_ID = DeckId(1)
DEFAULT_DECK_CONF_ID = DeckConfigId(1)


class DecksDictProxy:
    def __init__(self, col: anki.collection.Collection):
        self._col = col.weakref()

    def _warn(self) -> None:
        print_deprecation_warning(
            "add-on should use methods on col.decks, not col.decks.decks dict"
        )

    def __getitem__(self, item: Any) -> Any:
        self._warn()
        return self._col.decks.get(DeckId(int(item)))

    def __setitem__(self, key: Any, val: Any) -> None:
        self._warn()
        self._col.decks.save(val)

    def __len__(self) -> int:
        self._warn()
        return len(self._col.decks.all_names_and_ids())

    def keys(self) -> Any:
        self._warn()
        return [str(nt.id) for nt in self._col.decks.all_names_and_ids()]

    def values(self) -> Any:
        self._warn()
        return self._col.decks.all()

    def items(self) -> Any:
        self._warn()
        return [(str(nt["id"]), nt) for nt in self._col.decks.all()]

    def __contains__(self, item: Any) -> bool:
        self._warn()
        return self._col.decks.have(item)


class DeckManager(DeprecatedNamesMixin):
    # Registry save/load
    #############################################################

    def __init__(self, col: anki.collection.Collection) -> None:
        self.col = col.weakref()
        self.decks = DecksDictProxy(col)

    def save(self, deck_or_config: DeckDict | DeckConfigDict | None = None) -> None:
        "Can be called with either a deck or a deck configuration."
        if not deck_or_config:
            print("col.decks.save() should be passed the changed deck")
            return

        # deck conf?
        if "maxTaken" in deck_or_config:
            self.update_config(deck_or_config)
            return
        else:
            self.update(deck_or_config, preserve_usn=False)

    # Deck save/load
    #############################################################

    def add_normal_deck_with_name(self, name: str) -> OpChangesWithId:
        "If deck exists, return existing id."
        if id := self.col.decks.id_for_name(name):
            return OpChangesWithId(id=id)
        else:
            deck = self.col.decks.new_deck()
            deck.name = name
            return self.add_deck(deck)

    def add_deck_legacy(self, deck: DeckDict) -> OpChangesWithId:
        "Add a deck created with new_deck_legacy(). Must have id of 0."
        if not deck["id"] == 0:
            raise Exception("id should be 0")
        return self.col._backend.add_deck_legacy(to_json_bytes(deck))

    def id(
        self,
        name: str,
        create: bool = True,
        type: DeckConfigId = DeckConfigId(0),
    ) -> DeckId | None:
        "Add a deck with NAME. Reuse deck if already exists. Return id as int."
        id = self.id_for_name(name)
        if id:
            return id
        elif not create:
            return None

        deck = self.new_deck_legacy(bool(type))
        deck["name"] = name
        out = self.add_deck_legacy(deck)
        return DeckId(out.id)

    def remove(self, dids: Sequence[DeckId]) -> OpChangesWithCount:
        return self.col._backend.remove_decks(dids)

    def all_names_and_ids(
        self, skip_empty_default: bool = False, include_filtered: bool = True
    ) -> Sequence[DeckNameId]:
        "A sorted sequence of deck names and IDs."
        return self.col._backend.get_deck_names(
            skip_empty_default=skip_empty_default, include_filtered=include_filtered
        )

    def id_for_name(self, name: str) -> DeckId | None:
        try:
            return DeckId(self.col._backend.get_deck_id_by_name(name))
        except NotFoundError:
            return None

    def get_legacy(self, did: DeckId) -> DeckDict | None:
        try:
            return from_json_bytes(self.col._backend.get_deck_legacy(did))
        except NotFoundError:
            return None

    def have(self, id: DeckId) -> bool:
        return bool(self.get_legacy(id))

    def get_all_legacy(self) -> list[DeckDict]:
        return list(from_json_bytes(self.col._backend.get_all_decks_legacy()).values())

    def new_deck(self) -> Deck:
        "Return a new normal deck. It must be added with .add_deck() after a name assigned."
        return self.col._backend.new_deck()

    def add_deck(self, deck: Deck) -> OpChangesWithId:
        return self.col._backend.add_deck(message=deck)

    def new_deck_legacy(self, filtered: bool) -> DeckDict:
        deck = from_json_bytes(self.col._backend.new_deck_legacy(filtered))
        if deck["dyn"]:
            # Filtered decks are now created via a scheduler method, but old unit
            # tests still use this method. Set the default values to what the tests
            # expect: one empty search term, and ordering by oldest first.
            del deck["terms"][1]
            deck["terms"][0][0] = ""
            deck["terms"][0][2] = 0

        return deck

    def deck_tree(self) -> DeckTreeNode:
        return self.col._backend.deck_tree(now=0)

    @classmethod
    def find_deck_in_tree(
        cls, node: DeckTreeNode, deck_id: DeckId
    ) -> DeckTreeNode | None:
        if node.deck_id == deck_id:
            return node
        for child in node.children:
            match = cls.find_deck_in_tree(child, deck_id)
            if match:
                return match
        return None

    def all(self) -> list[DeckDict]:
        "All decks. Expensive; prefer all_names_and_ids()"
        return self.get_all_legacy()

    def set_collapsed(
        self, deck_id: DeckId, collapsed: bool, scope: DeckCollapseScope.V
    ) -> OpChanges:
        return self.col._backend.set_deck_collapsed(
            deck_id=deck_id, collapsed=collapsed, scope=scope
        )

    def collapse(self, did: DeckId) -> None:
        deck = self.get(did)
        deck["collapsed"] = not deck["collapsed"]
        self.save(deck)

    def collapse_browser(self, did: DeckId) -> None:
        deck = self.get(did)
        collapsed = deck.get("browserCollapsed", False)
        deck["browserCollapsed"] = not collapsed
        self.save(deck)

    def count(self) -> int:
        return len(self.all_names_and_ids())

    def card_count(
        self, dids: DeckId | Iterable[DeckId], include_subdecks: bool
    ) -> Any:
        if isinstance(dids, int):
            dids = {dids}
        else:
            dids = set(dids)
        if include_subdecks:
            dids.update([child[1] for did in dids for child in self.children(did)])
        str_ids = ids2str(dids)
        count = self.col.db.scalar(
            f"select count() from cards where did in {str_ids} or odid in {str_ids}"
        )
        return count

    def get(self, did: DeckId | str, default: bool = True) -> DeckDict | None:
        if not did:
            if default:
                return self.get_legacy(DEFAULT_DECK_ID)
            else:
                return None
        id = DeckId(int(did))
        deck = self.get_legacy(id)
        if deck:
            return deck
        elif default:
            return self.get_legacy(DEFAULT_DECK_ID)
        else:
            return None

    def by_name(self, name: str) -> DeckDict | None:
        """Get deck with NAME, ignoring case."""
        id = self.id_for_name(name)
        if id:
            return self.get_legacy(id)
        return None

    def update(self, deck: DeckDict, preserve_usn: bool = True) -> None:
        "Add or update an existing deck. Used for syncing and merging."
        deck["id"] = self.col._backend.add_or_update_deck_legacy(
            deck=to_json_bytes(deck), preserve_usn_and_mtime=preserve_usn
        )

    def update_dict(self, deck: DeckDict) -> OpChanges:
        return self.col._backend.update_deck_legacy(json=to_json_bytes(deck))

    def rename(self, deck: DeckDict | DeckId, new_name: str) -> OpChanges:
        "Rename deck prefix to NAME if not exists. Updates children."
        if isinstance(deck, int):
            deck_id = deck
        else:
            deck_id = deck["id"]
        return self.col._backend.rename_deck(deck_id=deck_id, new_name=new_name)

    # Drag/drop
    #############################################################

    def reparent(
        self, deck_ids: Sequence[DeckId], new_parent: DeckId
    ) -> OpChangesWithCount:
        """Rename one or more source decks that were dropped on `new_parent`.
        If new_parent is 0, decks will be placed at the top level."""
        return self.col._backend.reparent_decks(
            deck_ids=deck_ids, new_parent=new_parent
        )

    # Deck configurations
    #############################################################

    def get_deck_configs_for_update(self, deck_id: DeckId) -> DeckConfigsForUpdate:
        return self.col._backend.get_deck_configs_for_update(deck_id)

    def update_deck_configs(self, input: UpdateDeckConfigs) -> OpChanges:
        op_bytes = self.col._backend.update_deck_configs_raw(input.SerializeToString())
        return OpChanges.FromString(op_bytes)

    def all_config(self) -> list[DeckConfigDict]:
        "A list of all deck config."
        return list(from_json_bytes(self.col._backend.all_deck_config_legacy()))

    def config_dict_for_deck_id(self, did: DeckId) -> DeckConfigDict:
        deck = self.get(did, default=False)
        assert deck
        if "conf" in deck:
            dcid = DeckConfigId(int(deck["conf"]))  # may be a string
            conf = self.get_config(dcid)
            if not conf:
                # fall back on default
                conf = self.get_config(DEFAULT_DECK_CONF_ID)
            conf["dyn"] = False
            return conf
        # dynamic decks have embedded conf
        return deck

    def get_config(self, conf_id: DeckConfigId) -> DeckConfigDict | None:
        try:
            return from_json_bytes(self.col._backend.get_deck_config_legacy(conf_id))
        except NotFoundError:
            return None

    def update_config(self, conf: DeckConfigDict, preserve_usn: bool = False) -> None:
        "preserve_usn is ignored"
        conf["id"] = self.col._backend.add_or_update_deck_config_legacy(
            json=to_json_bytes(conf)
        )

    def add_config(
        self, name: str, clone_from: DeckConfigDict | None = None
    ) -> DeckConfigDict:
        if clone_from is not None:
            conf = copy.deepcopy(clone_from)
            conf["id"] = 0
        else:
            conf = from_json_bytes(self.col._backend.new_deck_config_legacy())
        conf["name"] = name
        self.update_config(conf)
        return conf

    def add_config_returning_id(
        self, name: str, clone_from: DeckConfigDict | None = None
    ) -> DeckConfigId:
        return self.add_config(name, clone_from)["id"]

    def remove_config(self, id: DeckConfigId) -> None:
        "Remove a configuration and update all decks using it."
        self.col.mod_schema(check=True)
        for deck in self.all():
            # ignore cram decks
            if "conf" not in deck:
                continue
            if str(deck["conf"]) == str(id):
                deck["conf"] = 1
                self.save(deck)
        self.col._backend.remove_deck_config(id)

    def set_config_id_for_deck_dict(self, deck: DeckDict, id: DeckConfigId) -> None:
        deck["conf"] = id
        self.save(deck)

    def decks_using_config(self, conf: DeckConfigDict) -> list[DeckId]:
        dids = []
        for deck in self.all():
            if "conf" in deck and deck["conf"] == conf["id"]:
                dids.append(deck["id"])
        return dids

    def restore_to_default(self, conf: DeckConfigDict) -> None:
        old_order = conf["new"]["order"]
        new = from_json_bytes(self.col._backend.new_deck_config_legacy())
        new["id"] = conf["id"]
        new["name"] = conf["name"]
        self.update_config(new)
        # if it was previously randomized, re-sort
        if not old_order:
            self.col.sched.resort_conf(new)

    # Deck utils
    #############################################################

    def name(self, did: DeckId, default: bool = False) -> str:
        deck = self.get(did, default=default)
        if deck:
            return deck["name"]
        return self.col.tr.decks_no_deck()

    def name_if_exists(self, did: DeckId) -> str | None:
        deck = self.get(did, default=False)
        if deck:
            return deck["name"]
        return None

    def cids(self, did: DeckId, children: bool = False) -> list[anki.cards.CardId]:
        if not children:
            return self.col.db.list("select id from cards where did=?", did)
        dids = [did]
        for name, id in self.children(did):
            dids.append(id)
        return self.col.db.list(f"select id from cards where did in {ids2str(dids)}")

    def for_card_ids(self, cids: list[anki.cards.CardId]) -> list[DeckId]:
        return self.col.db.list(f"select did from cards where id in {ids2str(cids)}")

    # Deck selection
    #############################################################

    def set_current(self, deck: DeckId) -> OpChanges:
        return self.col._backend.set_current_deck(deck)

    def get_current_id(self) -> DeckId:
        "The currently selected deck ID."
        return DeckId(self.col._backend.get_current_deck().id)

    def current(self) -> DeckDict:
        return self.get(self.selected())

    def active(self) -> list[DeckId]:
        # some add-ons assume this will always be non-empty
        return self.col.sched.active_decks or [DeckId(1)]

    def select(self, did: DeckId) -> None:
        # make sure arg is an int; legacy callers may be passing in a string
        did = DeckId(did)
        self.set_current(did)

    selected = get_current_id

    # Parents/children
    #############################################################

    @staticmethod
    def path(name: str) -> list[str]:
        return name.split("::")

    @classmethod
    def basename(cls, name: str) -> str:
        return cls.path(name)[-1]

    @classmethod
    def immediate_parent_path(cls, name: str) -> list[str]:
        return cls.path(name)[:-1]

    @classmethod
    def immediate_parent(cls, name: str) -> str | None:
        parent_path = cls.immediate_parent_path(name)
        if parent_path:
            return "::".join(parent_path)
        return None

    @classmethod
    def key(cls, deck: DeckDict) -> list[str]:
        return cls.path(deck["name"])

    def deck_and_child_name_ids(self, deck_id: DeckId) -> Iterable[tuple[str, DeckId]]:
        """The deck of did and all its children, as (name, id)."""
        return (
            (entry.name, DeckId(entry.id))
            for entry in self.col._backend.get_deck_and_child_names(deck_id)
        )

    def children(self, did: DeckId) -> list[tuple[str, DeckId]]:
        "All children of did, as (name, id)."
        return [
            name_id
            for name_id in self.deck_and_child_name_ids(did)
            if name_id[1] != did
        ]

    def child_ids(self, parent_name: str) -> Iterable[DeckId]:
        if not (parent_id := self.id_for_name(parent_name)):
            return []
        return (name_id[1] for name_id in self.children(parent_id))

    def deck_and_child_ids(self, deck_id: DeckId) -> list[DeckId]:
        return [
            DeckId(entry.id)
            for entry in self.col._backend.get_deck_and_child_names(deck_id)
        ]

    def parents(
        self, did: DeckId, name_map: dict[str, DeckDict] | None = None
    ) -> list[DeckDict]:
        "All parents of did."
        # get parent and grandparent names
        parents_names: list[str] = []
        for part in self.immediate_parent_path(self.get(did)["name"]):
            if not parents_names:
                parents_names.append(part)
            else:
                parents_names.append(f"{parents_names[-1]}::{part}")
        parents: list[DeckDict] = []
        # convert to objects
        for parent_name in parents_names:
            if name_map:
                deck = name_map[parent_name]
            else:
                deck = self.get(self.id(parent_name))
            parents.append(deck)
        return parents

    def parents_by_name(self, name: str) -> list[DeckDict]:
        "All existing parents of name"
        if "::" not in name:
            return []
        names = self.immediate_parent_path(name)
        head = []
        parents: list[DeckDict] = []

        while names:
            head.append(names.pop(0))
            deck = self.by_name("::".join(head))
            if deck:
                parents.append(deck)

        return parents

    # Filtered decks
    ##########################################################################

    def new_filtered(self, name: str) -> DeckId:
        "For new code, prefer col.sched.get_or_create_filtered_deck()."
        did = self.id(name, type=DEFAULT_DECK_CONF_ID)
        self.select(did)
        return did

    def is_filtered(self, did: DeckId | str) -> bool:
        return bool(self.get(did)["dyn"])

    # Legacy
    #############

    @deprecated(info="no longer required")
    def flush(self) -> None:
        pass

    @deprecated(replaced_by=remove)
    def rem(
        self,
        did: DeckId,
        **legacy_args: bool,
    ) -> None:
        "Remove the deck. If cardsToo, delete any cards inside."
        if isinstance(did, str):
            did = int(did)
        self.remove([did])

    @deprecated(replaced_by=all_names_and_ids)
    def name_map(self) -> dict[str, DeckDict]:
        return {d["name"]: d for d in self.all()}

    @deprecated(info="use col.set_deck() instead")
    def set_deck(self, cids: list[anki.cards.CardId], did: DeckId) -> None:
        self.col.set_deck(card_ids=cids, deck_id=did)
        self.col.db.execute(
            f"update cards set did=?,usn=?,mod=? where id in {ids2str(cids)}",
            did,
            self.col.usn(),
            int_time(),
        )

    @deprecated(replaced_by=all_names_and_ids)
    def all_ids(self) -> list[str]:
        return [str(x.id) for x in self.all_names_and_ids()]

    @deprecated(replaced_by=all_names_and_ids)
    def all_names(self, dyn: bool = True, force_default: bool = True) -> list[str]:
        return [
            x.name
            for x in self.all_names_and_ids(
                skip_empty_default=not force_default, include_filtered=dyn
            )
        ]


DeckManager.register_deprecated_aliases(
    confForDid=DeckManager.config_dict_for_deck_id,
    setConf=DeckManager.set_config_id_for_deck_dict,
    didsForConf=DeckManager.decks_using_config,
    allConf=DeckManager.all_config,
    getConf=DeckManager.get_config,
    updateConf=DeckManager.update_config,
    remConf=DeckManager.remove_config,
    confId=DeckManager.add_config_returning_id,
    newDyn=DeckManager.new_filtered,
    isDyn=DeckManager.is_filtered,
    nameOrNone=DeckManager.name_if_exists,
)


if not TYPE_CHECKING:

    def __getattr__(name):
        if name == "defaultDeck":
            print_deprecation_warning(
                "defaultDeck is deprecated; call decks.id() without it"
            )
            return 0
        elif name == "defaultDynamicDeck":
            print_deprecation_warning(
                "defaultDynamicDeck is replaced with new_filtered()"
            )
            return 1
        else:
            raise AttributeError(f"module {__name__} has no attribute {name}")
